# Copyright 2024 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Implementation of Engine API for MaxText"""
import functools
from typing import Any, List, Optional, Tuple, Callable
from collections import defaultdict
import uuid

import flax
from flax import linen as nn
from flax.linen import partitioning as nn_partitioning

from layers import models, quantizations

import jax
import jax.numpy as jnp
from jax.sharding import PartitionSpec as P
from jax.experimental import layout as jax_layout

import common_types
from jetstream.core import config_lib
from jetstream.engine import engine_api
from jetstream.engine import tokenizer_pb2
from jetstream.engine import tokenizer_api
from jetstream.engine import token_utils
from utils import lora_utils

import max_utils
import inference_utils
import pyconfig

import warnings

warnings.simplefilter("ignore", category=FutureWarning)
DecodeState = Any
Prefix = Any
PackedPrefix = Any
Params = Any
PRNGKeyType = Any
DLL = jax_layout.DeviceLocalLayout
Layout = jax_layout.Layout


class MaxEngineConfig:
  """Engine specific config class to allow using multiple MaxEngine instances in an inference run.
  TODO: evaluate the need for this given the restructured pyconfig.py
  """

  def __init__(self, keys):
    # self.keys = keys
    self.__dict__["keys"] = keys

  def __getattr__(self, attr):
    if attr not in self.keys:
      raise ValueError(f"Requested key {attr}, not in config")
    return self.keys[attr]

  def __setattr__(self, attr, value):
    raise ValueError

  def get_keys(self):
    return self.keys


class MaxEngine(engine_api.Engine):
  """The computational core of the generative model server.

  Engine defines an API that models must adhere to as they plug into the
  JetStream efficient serving infrastructure.
  """

  def __init__(self, config: Any, devices: config_lib.Devices | None = None):
    self.config = config

    # Mesh definition
    devices_array = max_utils.create_device_mesh(config=config, devices=devices)
    self._mesh = jax.sharding.Mesh(devices_array, config.mesh_axes)

    # Model and Optimizer definition
    quant = quantizations.configure_quantization(config)
    self.model = models.Transformer(config, mesh=self._mesh, quant=quant)
    self.replicated_sharding = jax.sharding.NamedSharding(self._mesh, P(None))

    self.abstract_params = None
    self.prefill_kv_cache_annotations = None
    self.kv_cache_annotations = None
    self.kv_cache_annotations_named = None
    self.prefill_kv_cache_shardings = None
    self.kv_cache_shardings = None
    self.state_mesh_annotations = None
    self.decode_state_shapes = None
    self.decode_state_layouts = None
    self.param_layouts = None

  def print_stats(self, label: str):
    max_utils.print_mem_stats(label)
    max_utils.print_cpu_ram_stats(label)

  def generate_aot(
      self, params: Params, decode_state: DecodeState, rng: Optional[PRNGKeyType] = None
  ) -> Tuple[DecodeState, engine_api.ResultTokens]:
    """Wrapper to generate for ahead of time compilation."""

    return self.generate(params=params, decode_state=decode_state, rng=rng)

  def _compile_generate_and_get_layouts(
      self, params: Any, decode_state: Any, rng_shape: Any, xla_flags: dict[str, Any] | None = None
  ) -> tuple[Any, Any, Any, Any]:
    """Optimal memory layout for params and decode_state."""

    param_layout = Layout(DLL.AUTO)
    decode_state_layout = Layout(DLL.AUTO)
    # Keyword arguments are not yet supported in JAX for specifying shardings. Therefore, all AOT
    # compiled functions use arguments instead.
    compiled_generate = (
        jax.jit(
            self.generate_aot,
            in_shardings=(param_layout, decode_state_layout, None),
            out_shardings=(Layout(DLL.AUTO), Layout(DLL.AUTO)),
            donate_argnames=("decode_state",),
        ).lower(params, decode_state, rng_shape)
    ).compile(compiler_options=xla_flags)

    arg_layouts, _ = compiled_generate.input_layouts
    generate_out_layouts, _ = compiled_generate.output_layouts

    return compiled_generate, arg_layouts[0], arg_layouts[1], generate_out_layouts

  def _identity(self, x: Any) -> Any:
    """Avoids lambda that breaks JAX caching."""

    return x

  def _iterated_layout(self, arrays: Any, layouts: Any, xla_flags: dict[str, Any] | None = None) -> Any:
    """Lays out an array tensor by tensor to prevent OOMs."""

    def _layout(x, s, l):
      if x.layout == l:
        return x
      # Somehow this can be None sometimes.
      dll = l.device_local_layout if isinstance(l, Layout) else l
      f = jax.jit(self._identity, out_shardings=Layout(dll, s)).lower(x).compile(compiler_options=xla_flags)
      y = f(x)
      # Achieves donation of the input argument, but allows for different memory
      # layouts and shapes.
      jax.tree.map(lambda z: z.delete(), x)
      jax.block_until_ready(y)
      return y

    shardings = jax.tree.map(lambda x: x.sharding, arrays)
    arrays = jax.tree.map(_layout, arrays, shardings, layouts)
    return arrays

  def aot_compile(
      self, params: Params, pass_rng_shape: bool, xla_flags: dict[str, Any] | None = None
  ) -> Tuple[Any, Params, Any]:
    """Ahead of time compilation of generate with auto layout, relayout parameters."""
    if pass_rng_shape:
      rng_shape = jax.ShapeDtypeStruct([4], jax.numpy.dtype("uint32"))
    else:
      rng_shape = None
    self.decode_state_shapes = jax.eval_shape(self.init_decode_state, rng_shape)

    generate_executable, self.param_layouts, _, self.decode_state_layouts = self._compile_generate_and_get_layouts(
        self.abstract_params, self.decode_state_shapes, rng_shape, xla_flags
    )
    return (
        generate_executable,
        self._iterated_layout(params, self.param_layouts),
        jax.jit(self.init_decode_state, in_shardings=(None), out_shardings=self.decode_state_layouts)
        .lower(rng_shape)
        .compile(),
    )

  def load_params(self, *args, params=None, rng: Optional[PRNGKeyType] = None, **kwargs) -> Params:
    """Load Parameters from GCS or reshard given Parameters"""
    # pylint: disable=unused-argument

    if rng is None:
      rng = jax.random.PRNGKey(0)

    if self.model.quant and self.config.checkpoint_is_quantized:
      print("Loading from the quantized checkpoint...")
      self.model.quant.quant_mode = quantizations.get_quant_mode("serve")

    rng1, rng2, rng3 = jax.random.split(rng, 3)
    if params:
      print("Resharding given params")
      _, self.state_mesh_annotations, state_mesh_shardings = max_utils.get_abstract_state(
          self.model, None, self.config, rng, self._mesh, False
      )
      # reshard given params based on shardings from config in MaxEngine
      params = jax.device_put(params, state_mesh_shardings.params)
      state = max_utils.init_decode_state(None, params)
      state = max_utils.unbox_logicallypartioned(state)
    else:
      state, self.state_mesh_annotations = max_utils.setup_decode_state(self.model, self.config, rng1, self._mesh, None)
    # pylint: disable=isinstance-second-argument-not-valid-type
    self.abstract_params = jax.tree_util.tree_map(
        lambda x: jax.ShapeDtypeStruct(shape=x.shape, dtype=x.dtype, sharding=x.sharding)
        if isinstance(x, jax.Array)
        else None,
        state.params,
    )

    self.prefill_kv_cache_annotations = max_utils.get_prefill_kv_cache_annotations(self.model, self.config, rng2, self._mesh)
    self.prefill_kv_cache_shardings = jax.tree_util.tree_map(
        lambda x: jax.sharding.NamedSharding(self._mesh, x),
        self.prefill_kv_cache_annotations,
    )

    if self.config.stack_prefill_result_cache:
      # Add extra axis for the axis generated by the stack.
      self.prefill_kv_cache_shardings = jax.tree_util.tree_map(
          lambda x: jax.sharding.NamedSharding(self._mesh, jax.sharding.PartitionSpec(None, *x.spec)),
          self.prefill_kv_cache_shardings,
      )
      self.prefill_kv_cache_shardings = self.prefill_kv_cache_shardings["decoder"]["layers_0"]

    self.kv_cache_annotations = max_utils.get_kv_cache_annotations(self.model, self.config, rng2, self._mesh)
    self.kv_cache_shardings = jax.tree_util.tree_map(
        lambda x: jax.sharding.NamedSharding(self._mesh, x),
        self.kv_cache_annotations,
    )

    if self.model.quant and not self.config.checkpoint_is_quantized:
      params = self.quantize_params(state, rng3)
    else:
      params = state.params

    self.print_stats("After load_params")

    return params

  def load_single_adapter(self, adapter_path):
    """
    Load Single adapter from adapter_path.
    Expect adapter_config.json and LoRA adapter weights at this path within subdirectory `/0/items`.
    """
    adapter_config_path = f"{adapter_path}/adapter_config.json"
    adapter_weights_path = f"{adapter_path}/0/items"

    params, config = lora_utils.load_adapter(self.config, self.abstract_params, adapter_config_path, adapter_weights_path)

    if config is None:
      raise ValueError(f"Failed to read lora_config from {adapter_config_path}")

    if params is None:
      raise ValueError(f"Failed to read lora_config from {adapter_config_path}")

    config["adapter_path"] = adapter_weights_path

    self.print_stats("After load_single_adapter.")

    return params, config

  def apply_adapter(self, base_params, adapter_config, adapter_params):
    """Apply the adapter params on the base params."""

    lora_rank = int(adapter_config["r"])
    lora_scale_factor = float(adapter_config["lora_alpha"]) / lora_rank
    lora_utils.apply_lora_on_base_params(base_params, adapter_params, lora_scale_factor)

  def unapply_adapter(self, base_params, adapter_config, adapter_params):
    """Unapply the adapter params from the merged params to get back the base params."""

    lora_rank = int(adapter_config["r"])
    lora_scale_factor = float(adapter_config["lora_alpha"]) / lora_rank
    lora_utils.unapply_lora_from_base_params(base_params, adapter_params, lora_scale_factor)

  def quantize_params(self, state, rng: Optional[PRNGKeyType] = None):
    """Forward pass to quantize decode params."""
    if rng is None:
      rng = jax.random.PRNGKey(0)

    self.model.quant.quant_mode = quantizations.get_quant_mode("convert")

    @jax.jit
    def model_apply(_p, _rng):
      return self.model.apply(
          _p | {"aqt": {}},
          jnp.ones((1, self.config.max_prefill_predict_length), dtype=jnp.int32),
          jnp.ones((1, self.config.max_prefill_predict_length), dtype=jnp.int32),
          decoder_segment_ids=jnp.zeros((1, self.config.max_prefill_predict_length), dtype=jnp.int32),
          enable_dropout=False,
          model_mode=common_types.MODEL_MODE_PREFILL,
          rngs={"params": _rng},
          mutable=True,
      )

    _, new_vars = model_apply(state.params, rng)
    # Remove param values which have corresponding qtensors in aqt to save memory.
    params = {}
    params["aqt"] = new_vars["aqt"]
    params["params"] = quantizations.remove_quantized_params(state.params["params"], new_vars["aqt"])
    self.abstract_params = jax.tree_util.tree_map(
        lambda x: jax.ShapeDtypeStruct(shape=x.shape, dtype=x.dtype, sharding=x.sharding),
        params,
    )
    max_utils.save_quantized_checkpoint_if_configured(self.config, params)
    self.model.quant.quant_mode = quantizations.get_quant_mode("serve")
    return params

  def _maybe_stack_prefill_result_cache(self, cache):
    """Stack the caches across the layers."""
    if not self.config.stack_prefill_result_cache:
      return cache

    layer_keys = []
    for i in range(self.config.num_decoder_layers):
      layer_keys.append(f"layers_{i}")

    layer_cache = [cache["decoder"][layer_key] for layer_key in layer_keys]

    return jax.tree.map(lambda *c: jnp.stack(c), *layer_cache)

  def _maybe_unstack_prefill_result_cache(self, cache):
    """Unstack the caches across the layers."""
    if not self.config.stack_prefill_result_cache:
      return cache

    flat_cache, treedef = jax.tree.flatten(cache)
    layer_cache = [jax.tree.unflatten(treedef, flat_cache_vars) for flat_cache_vars in zip(*flat_cache, strict=True)]
    res_cache = {"decoder": {}}

    for i in range(self.config.num_decoder_layers):
      res_cache["decoder"][f"layers_{i}"] = layer_cache[i]

    return res_cache

  def prefill_aot(  # pylint: disable=too-many-positional-arguments
      self,
      params: Params,
      padded_tokens: jax.Array,
      true_length: int,
      rng: Optional[PRNGKeyType] = None,
      complete_prompt_true_length: Optional[int] = None,
      complete_padded_prompt: Optional[jax.Array] = None,
      positions: Optional[jax.Array] = None,
      previous_chunk: Optional[Any] = None,
  ) -> Tuple[Prefix, engine_api.ResultTokens]:
    """Wrapper for prefill for ahead-of-time compilation."""

    return self.prefill(
        params=params,
        padded_tokens=padded_tokens,
        true_length=true_length,
        rng=rng,
        complete_prompt_true_length=complete_prompt_true_length,
        complete_padded_prompt=complete_padded_prompt,
        positions=positions,
        previous_chunk=previous_chunk,
    )

  @functools.partial(jax.jit, static_argnums=(0,), static_argnames=("request_id",))
  def prefill(
      self,
      *,
      params: Params,
      existing_prefix: Optional[jax.Array] = None,
      padded_tokens: jax.Array,
      true_length: int,
      sampler: Optional[Callable[[Any], Any]] = None,  # pylint: disable=unused-argument
      rng: Optional[PRNGKeyType] = None,
      complete_prompt_true_length: Optional[int] = None,
      complete_padded_prompt: Optional[jax.Array] = None,
      positions: Optional[jax.Array] = None,
      previous_chunk: Optional[Any] = None,
      request_id: Optional[uuid.UUID] = None,  # pylint: disable=unused-argument
      slot: Optional[int] = None,
  ) -> Tuple[Prefix, engine_api.ResultTokens]:
    """Computes a kv-cache for a new generate request.

    Args:
      params: Scalar multiplier.
      existing_prefix: If provided, represents a prefix that has already been
        processed by the underlying model.

      padded_tokens: Logically appended tokens to any existing prefix, this is
        what we compute prefill on.
        If chunked prefill is true - this represents current padded chunk. (eg the last chunk might need to be padded)
        else - complete padded prompt

      true_length: The real length of the tokens, pre-pad.
                  If chunked prefill is true - this represents true length of the current chunk
                  else - true length of complete prompt
      complete_prompt_true_length: true length of the entire prompt
                  it can be none if chunked prefill is false
                  if chunked prefill is true, it is needed for constructing final decoder segment Ids.
      complete_padded_prompt: Optional[jax.Array] = None,
                   it can be none if chunked prefill is false
                   if chunked prefill is true, it is needed for constructing decoder active sequence indicator
      positions: Optional[jax.Array] = None,
                current position of the tokens in chunk - used for rope embeddings
      previous_chunk: Optional[Any] = None, - Has relevant information from previous processed chunks
      mainly - next postion and KV cache

    relevant params in call for chunked prefill where complete length is 10 (12 after padding),
    chunk size is 4 and current chunk is second chunk

    padded_tokens = [t4, t5, t6, t7]
    true_length = 4
    complete_prompt_true_length = 10
    complete_padded_prompt = [t0, t1, t2, t3...t11]
    positions = [4, 5, 6, 7]
    previous_chunk = {'cache':{}}

    If chunked prefill is false and prefill call is for entire prompt
    padded_tokens = [t0, t1, .., t11]
    true_length = 10
    complete_prompt_true_length = None
    complete_padded_prompt = None
    positions = None
    previous_chunk = None

    Returns:
      kv_cache: For the resulting text.
    """
    if existing_prefix:
      raise ValueError("We don't know what to do with existing_prefix")

    if rng is None:
      rng = jax.random.PRNGKey(0)

    input_tokens = jnp.expand_dims(padded_tokens, 0)  # [BATCH, SEQUENCE]
    if positions is None:
      positions = jnp.expand_dims(jnp.arange(0, input_tokens.shape[1]), 0)
    if not self.config.use_chunked_prefill:
      zero_to_n = jnp.arange(0, padded_tokens.shape[0])
      ones_to_keep = zero_to_n < true_length
      next_pos = jnp.full((1, 1), true_length, dtype=jnp.int32)
    else:
      zero_to_n = jnp.arange(0, complete_padded_prompt.shape[0])  # pytype: disable=attribute-error
      ones_to_keep = zero_to_n < complete_prompt_true_length
      next_pos = jnp.full((1, 1), complete_prompt_true_length, dtype=jnp.int32)

    one_d_output = ones_to_keep * common_types.DECODING_ACTIVE_SEQUENCE_INDICATOR
    sequence_indicator = jnp.expand_dims(one_d_output, 0)

    rng, new_rng = jax.random.split(rng)
    with self._mesh, nn_partitioning.axis_rules(self.config.logical_axis_rules):
      flat_logits, new_vars = self.model.apply(
          params,
          input_tokens,
          positions,
          decoder_segment_ids=sequence_indicator,
          enable_dropout=False,
          model_mode=common_types.MODEL_MODE_PREFILL,
          rngs={"params": new_rng},
          mutable=["cache"],
          previous_chunk=previous_chunk,
          true_length=true_length,
          slot=slot,
      )
    generated_tokens = jnp.zeros((1, 1), dtype=jnp.int32)
    selected_logits = jax.lax.dynamic_slice(
        flat_logits,
        (0, true_length - 1, 0),
        (flat_logits.shape[0], 1, flat_logits.shape[2]),
    )
    selected_logits = jax.lax.with_sharding_constraint(selected_logits, self.replicated_sharding)

    # sampling first token
    first_generated_token = inference_utils.sampling(
        selected_logits,
        rng,
        self.config.decode_sampling_strategy,
        topk=self.config.decode_sampling_top_k,
        nucleus_topp=self.config.decode_sampling_nucleus_p,
        temperature=self.config.decode_sampling_temperature,
    )

    all_valid = jnp.ones(first_generated_token.shape, dtype=jnp.int8)
    result = engine_api.ResultTokens(
        data=jnp.concatenate((first_generated_token, all_valid, generated_tokens), axis=1),
        # Tokens are shape [batch, speculations], so when we concatenate
        # tokens, validity and length along their index 1 dimension then they
        # occupy 0:speculations.
        tokens_idx=(0, 1),
        # Validity occupies the same amount of space, but next in line.
        valid_idx=(1, 2),
        # And lengths is rank 1.
        length_idx=(2, 3),
        samples_per_slot=1,
    )

    cache = new_vars["cache"]
    cache = self._maybe_stack_prefill_result_cache(cache)
    return {
        "logits": selected_logits,
        "cache": cache,
        "next_pos": next_pos,
        "generated_tokens": generated_tokens,
        "tokens": first_generated_token,
    }, result

  def prefill_multisampling_aot(  # pylint: disable=too-many-positional-arguments
      self,
      params: Params,
      padded_tokens: jax.Array,
      true_length: int,
      rng: Optional[PRNGKeyType] = None,
      num_samples: int = 1,
      sampler: Optional[Callable[[Any], Any]] = None,
  ) -> Tuple[Prefix, engine_api.ResultTokens]:
    """Wrapper for multi-sampling prefill for ahead-of-time compilation."""
    return self.prefill_multisampling(
        params=params,
        padded_tokens=padded_tokens,
        true_length=true_length,
        sampler=sampler,
        rng=rng,
        num_samples=num_samples,
    )

  @functools.partial(jax.jit, static_argnums=(0,), static_argnames=("num_samples",))
  def prefill_multisampling(
      self,
      *,
      params: Params,
      padded_tokens: jax.Array,
      true_length: int,
      sampler: Optional[Callable[[Any], Any]] = None,  # pylint: disable=unused-argument
      rng: Optional[PRNGKeyType] = None,
      num_samples: int = 1,
  ) -> Tuple[Prefix, engine_api.ResultTokens]:
    """Computes a kv-cache for a new generate request.

    With multi-sampling, the engine will generate multiple first tokens in the
    prefilling stage. The number of tokens is specified by num_samples.
    """

    if rng is None:
      rng = jax.random.PRNGKey(0)

    input_tokens = jnp.expand_dims(padded_tokens, 0)  # [BATCH, SEQUENCE]
    positions = jnp.expand_dims(jnp.arange(0, input_tokens.shape[1]), 0)

    zero_to_n = jnp.arange(0, padded_tokens.shape[0])
    ones_to_keep = zero_to_n < true_length
    one_d_output = ones_to_keep * common_types.DECODING_ACTIVE_SEQUENCE_INDICATOR
    sequence_indicator = jnp.expand_dims(one_d_output, 0)

    rng, new_rng = jax.random.split(rng)
    with self._mesh, nn_partitioning.axis_rules(self.config.logical_axis_rules):
      flat_logits, new_vars = self.model.apply(
          params,
          input_tokens,
          positions,
          decoder_segment_ids=sequence_indicator,
          enable_dropout=False,
          model_mode=common_types.MODEL_MODE_PREFILL,
          rngs={"params": new_rng},
          mutable=["cache"],
      )

    next_pos = jnp.full((1, 1), true_length, dtype=jnp.int32)
    selected_logits = jax.lax.dynamic_slice(
        flat_logits,
        (0, true_length - 1, 0),
        (flat_logits.shape[0], 1, flat_logits.shape[2]),
    )
    selected_logits = jax.lax.with_sharding_constraint(selected_logits, self.replicated_sharding)

    # sampling first tokens
    first_generated_tokens = []
    for _ in range(num_samples):
      rng, new_rng = jax.random.split(rng)
      first_generated_token = inference_utils.sampling(
          selected_logits,
          new_rng,
          self.config.decode_sampling_strategy,
          topk=self.config.decode_sampling_top_k,
          nucleus_topp=self.config.decode_sampling_nucleus_p,
          temperature=self.config.decode_sampling_temperature,
      )
      first_generated_tokens.append(first_generated_token)
    first_generated_tokens = jnp.concatenate(first_generated_tokens, axis=0)

    all_valid = jnp.ones((num_samples, 1), dtype=jnp.int8)
    generated_tokens = jnp.zeros((num_samples, 1), dtype=jnp.int32)
    result = engine_api.ResultTokens(
        data=jnp.concatenate((first_generated_tokens, all_valid, generated_tokens), axis=1),
        # Tokens are shape [batch, speculations], so when we concatenate
        # tokens, validity and length along their index 1 dimension then they
        # occupy 0:speculations.
        tokens_idx=(0, 1),
        # Validity occupies the same amount of space, but next in line.
        valid_idx=(1, 2),
        # And lengths is rank 1.
        length_idx=(2, 3),
        samples_per_slot=num_samples,
    )

    cache = new_vars["cache"]
    cache = self._maybe_stack_prefill_result_cache(cache)

    return {
        "logits": selected_logits,
        "cache": cache,
        "next_pos": next_pos,
        "generated_tokens": generated_tokens,
        "tokens": first_generated_tokens,
    }, result

  @functools.partial(jax.jit, static_argnums=(0,), static_argnames=("num_prompts",))
  def prefill_concat(
      self,
      *,
      params: Params,
      existing_prefix: Optional[jax.Array] = None,
      padded_tokens: jax.Array,
      decoder_positions: jax.Array,
      decoder_segment_ids: jax.Array,
      start_pos: jax.Array,
      true_lengths: jax.Array,
      num_prompts: int,
      sampler: Optional[Callable[[Any], Any]] = None,  # pylint: disable=unused-argument
      rng: Optional[PRNGKeyType] = None,
  ) -> Tuple[Any, PackedPrefix, List[engine_api.ResultTokens]]:
    """Computes a kv-cache for a new packed generate request, which is a
    concatenation of several shorter prompts. Experimentation shows that
    longer prefill sequences gives approximately 15% boost in time per prefilled
    token.

    Args:
      params: Scalar multiplier.
      existing_prefix: If provided, represents a prefix that has already been
        processed by the underlying model.
      padded_tokens: Logically appended tokens to any existing prefix, this is
        what we compute prefill on.
      decoder_positions: int values indicating the position of token in its
        original sequence.
      decoder_segment_ids: int values indicating which sequence the the token
        originally belong to.
      start_pos: Padded array indicating the start position of each of the prompts.
      true_length: Padded array indicating the true lengths of each of the prompts.
      num_prompts: the number of prompts packed in the entire sequence.
    Returns:
      kv_cache: For the resulting text.
    """
    if existing_prefix:
      raise ValueError("We don't know what to do with existing_prefix")

    if rng is None:
      rng = jax.random.PRNGKey(0)
    input_tokens = jnp.expand_dims(padded_tokens, 0)  # [BATCH, SEQUENCE]
    decoder_positions = jnp.expand_dims(decoder_positions, 0)
    decoder_segment_ids = jnp.expand_dims(decoder_segment_ids, 0)
    rng, new_rng = jax.random.split(rng)
    with self._mesh, nn_partitioning.axis_rules(self.config.logical_axis_rules):
      flat_logits, new_vars = self.model.apply(
          params,
          input_tokens,
          decoder_positions,
          decoder_segment_ids=decoder_segment_ids,
          enable_dropout=False,
          model_mode=common_types.MODEL_MODE_PREFILL,
          rngs={"params": new_rng},
          mutable=["cache"],
      )
    cache = new_vars["cache"]
    cache = self._maybe_stack_prefill_result_cache(cache)

    def process_packed_logits_and_caches(packed_flat_logits, idx):
      next_pos = jnp.full((1, 1), true_lengths[idx], dtype=jnp.int32)
      generated_tokens = jnp.zeros((1, 1), dtype=jnp.int32)
      selected_logits = jax.lax.dynamic_slice(
          packed_flat_logits,
          (0, start_pos[idx] + true_lengths[idx] - 1, 0),
          (packed_flat_logits.shape[0], 1, packed_flat_logits.shape[2]),
      )
      selected_logits = jax.lax.with_sharding_constraint(selected_logits, self.replicated_sharding)
      first_generated_token = inference_utils.sampling(
          selected_logits,
          rng,
          self.config.decode_sampling_strategy,
          topk=self.config.decode_sampling_top_k,
          nucleus_topp=self.config.decode_sampling_nucleus_p,
          temperature=self.config.decode_sampling_temperature,
      )
      all_valid = jnp.ones(first_generated_token.shape, dtype=jnp.int8)
      result = engine_api.ResultTokens(
          data=jnp.concatenate((first_generated_token, all_valid, generated_tokens), axis=1),
          # Tokens are shape [batch, speculations], so when we concatenate
          # tokens, validity and length along their index 1 dimension then they
          # occupy 0:speculations.
          tokens_idx=(0, 1),
          # Validity occupies the same amount of space, but next in line.
          valid_idx=(1, 2),
          # And lengths is rank 1.
          length_idx=(2, 3),
          samples_per_slot=1,
      )
      return {
          "logits": selected_logits,
          "next_pos": next_pos,
          "generated_tokens": generated_tokens,
          "tokens": first_generated_token,
      }, result

    prefill_results = defaultdict(list)
    first_tokens = []
    for idx in range(num_prompts):
      prefill_result, first_token = process_packed_logits_and_caches(flat_logits, idx)
      for k, v in prefill_result.items():
        prefill_results[k].append(v)
      first_tokens.append(first_token)
    prefill_results = {k: jnp.stack(v) for k, v in prefill_results.items()}
    return cache, prefill_results, first_tokens

  def prefill_insert(  # pylint: disable=too-many-positional-arguments
      self,
      padded_tokens: jax.Array,
      true_length: int,
      rng: Any,
      decode_state: DecodeState,
      slot: int,
      params: Params,
  ) -> DecodeState:
    """Prefill and insert a single computed prefill cache into KV cache."""

    prefix, _ = self.prefill(params=params, padded_tokens=padded_tokens, true_length=true_length, rng=rng)
    return self.insert(prefix, decode_state, slot)

  @functools.partial(jax.jit, static_argnums=(0,), donate_argnums=(2,))
  def generate(
      self,
      params: Params,
      decode_state: DecodeState,
      sampler: Optional[Callable[[Any], Any]] = None,  # pylint: disable=unused-argument
      rng: Optional[PRNGKeyType] = None,
  ) -> Tuple[DecodeState, engine_api.ResultTokens]:
    """Run one generate step"""
    if rng is None:
      rng = jax.random.PRNGKey(0)

    previous_token = decode_state["tokens"]
    rng, new_rng = jax.random.split(rng)
    # run one step generation
    with self._mesh, nn_partitioning.axis_rules(self.config.logical_axis_rules):
      out_logits, new_vars = self.model.apply(
          params | {"cache": decode_state["cache"]},
          previous_token,
          decode_state["next_pos"],
          enable_dropout=False,
          model_mode=common_types.MODEL_MODE_AUTOREGRESSIVE,
          rngs={"params": new_rng},
          mutable=["cache"],
      )
    out_logits = jax.lax.with_sharding_constraint(out_logits, self.replicated_sharding)
    new_cache = jax.lax.with_sharding_constraint(new_vars["cache"], self.kv_cache_shardings)
    # sampling tokens
    new_token = inference_utils.sampling(
        out_logits,
        rng,
        self.config.decode_sampling_strategy,
        topk=self.config.decode_sampling_top_k,
        nucleus_topp=self.config.decode_sampling_nucleus_p,
        temperature=self.config.decode_sampling_temperature,
    )
    all_valid = jnp.ones(new_token.shape, dtype=jnp.int8)
    result = engine_api.ResultTokens(
        data=jnp.concatenate((new_token, all_valid, decode_state["generated_tokens"]), axis=1),
        # Tokens are shape [batch, speculations], so when we concatenate
        # tokens, validity and length along their index 1 dimension then they
        # occupy 0:speculations.
        tokens_idx=(0, 1),
        # Validity occupies the same amount of space, but next in line.
        valid_idx=(1, 2),
        # And lengths is rank 1.
        length_idx=(2, 3),
        samples_per_slot=1,
    )

    return {
        "logits": out_logits,
        "cache": new_cache,
        "next_pos": decode_state["next_pos"] + 1,
        "generated_tokens": decode_state["generated_tokens"] + 1,
        "tokens": new_token,
    }, result

  @functools.partial(
      jax.jit,
      static_argnums=(0,),
      donate_argnums=(
          1,
          2,
      ),
  )
  def bulk_insert(
      self,
      prefix: Prefix,
      decode_state: DecodeState,
      slots: list[int],
  ) -> DecodeState:
    """Insert a single computed prefill cache into multiple slots in KV cache."""
    unboxed_prefix = max_utils.unbox_logicallypartioned(prefix)

    unboxed_prefix["cache"] = self._maybe_unstack_prefill_result_cache(unboxed_prefix["cache"])

    def copy(path, partial_cache, full_cache, annotations):
      path_key = path[-1].key
      if path_key in [
          "cache_ar_index",
          "cached_ar_key",
          "cached_ar_value",
          "cached_ar_key_scale",
          "cached_ar_value_scale",
      ]:
        return full_cache  # we don't even zero these out because we can mask them out.

      batch_idx = -1
      if "cache_batch" in annotations:
        batch_idx = annotations.index("cache_batch")
      elif "cache_scale_batch" in annotations:
        batch_idx = annotations.index("cache_scale_batch")

      if batch_idx < 0:
        raise ValueError(f"Batch index {batch_idx=} shouldn't be less than zero for {path_key}, got {annotations=}")

      for slot in slots:
        if path_key == "cache_ar_segment_id":
          ### goal: zero this out in case there is existing data
          s = list(full_cache.shape)
          s[batch_idx] = 1
          zeros = jnp.zeros(tuple(s), dtype=jnp.int32)
          full_cache = jax.lax.dynamic_update_index_in_dim(full_cache, zeros, slot, batch_idx)
        elif path_key == "cache_prefill_segment_id":
          s = list(full_cache.shape)
          s[batch_idx] = 1
          zeros = jnp.zeros(tuple(s), dtype=jnp.int32)
          ## zero out in case prefill cache is too small to cover
          full_cache = jax.lax.dynamic_update_index_in_dim(full_cache, zeros, slot, batch_idx)
          ## copy prefill cachce
          full_cache = jax.lax.dynamic_update_index_in_dim(full_cache, partial_cache, slot, batch_idx)
        elif path_key == "cached_ar_lengths":
          full_cache = full_cache.at[slot].set(0)
        elif path_key in [
            "cached_prefill_key",
            "cached_prefill_value",
            "cached_prefill_key_scale",
            "cached_prefill_value_scale",
        ]:
          full_cache = jax.lax.dynamic_update_index_in_dim(full_cache, partial_cache, slot, batch_idx)
        else:
          raise ValueError(f"We don't have a strategy for inserting {path_key}")

      return full_cache

    inserted_cache = jax.tree_util.tree_map_with_path(
        copy,
        unboxed_prefix["cache"],
        decode_state["cache"],
        self.kv_cache_annotations_named,
    )

    for i, slot in enumerate(slots):
      decode_state["logits"] = jax.lax.dynamic_update_index_in_dim(decode_state["logits"], unboxed_prefix["logits"], slot, 0)
      decode_state["next_pos"] = jax.lax.dynamic_update_index_in_dim(
          decode_state["next_pos"], unboxed_prefix["next_pos"], slot, 0
      )
      decode_state["generated_tokens"] = jax.lax.dynamic_update_index_in_dim(
          decode_state["generated_tokens"],
          jnp.expand_dims(unboxed_prefix["generated_tokens"][i], axis=0),
          slot,
          0,
      )
      decode_state["tokens"] = jax.lax.dynamic_update_index_in_dim(
          decode_state["tokens"],
          jnp.expand_dims(unboxed_prefix["tokens"][i], axis=0),
          slot,
          0,
      )

    inserted_logits = jax.lax.with_sharding_constraint(decode_state["logits"], self.replicated_sharding)
    inserted_generated_tokens = jax.lax.with_sharding_constraint(decode_state["generated_tokens"], self.replicated_sharding)
    inserted_next_pos = jax.lax.with_sharding_constraint(decode_state["next_pos"], self.replicated_sharding)
    inserted_tokens = jax.lax.with_sharding_constraint(decode_state["tokens"], self.replicated_sharding)
    inserted_cache = jax.lax.with_sharding_constraint(inserted_cache, self.kv_cache_shardings)

    return {
        "logits": inserted_logits,
        "cache": inserted_cache,
        "next_pos": inserted_next_pos,
        "generated_tokens": inserted_generated_tokens,
        "tokens": inserted_tokens,
    }

  @functools.partial(
      jax.jit,
      static_argnums=(0,),
      donate_argnums=(
          1,
          2,
      ),
      static_argnames=("request_id",),
  )
  def insert(
      self,
      prefix: Prefix,
      decode_state: DecodeState,
      slot: int,
      request_id: Optional[uuid.UUID] = None,  # pylint: disable=unused-argument
  ) -> DecodeState:
    """Insert a single computed prefill cache into KV cache."""
    unboxed_prefix = max_utils.unbox_logicallypartioned(prefix)

    unboxed_prefix["cache"] = self._maybe_unstack_prefill_result_cache(unboxed_prefix["cache"])

    def copy(path, partial_cache, full_cache, annotations):
      path_key = path[-1].key
      if path_key in [
          "cache_ar_index",
          "cached_ar_key",
          "cached_ar_value",
          "cached_ar_key_scale",
          "cached_ar_value_scale",
      ]:
        return full_cache  # we don't even zero these out because we can mask them out.

      batch_idx = -1
      if "cache_batch" in annotations:
        batch_idx = annotations.index("cache_batch")
      elif "cache_scale_batch" in annotations:
        batch_idx = annotations.index("cache_scale_batch")

      if batch_idx < 0:
        raise ValueError(f"Batch index {batch_idx=} shouldn't be less than zero for {path_key}, got {annotations=}")

      if path_key == "cache_ar_segment_id":
        ### goal: zero this out in case there is existing data
        s = list(full_cache.shape)
        s[batch_idx] = 1
        zeros = jnp.zeros(tuple(s), dtype=jnp.int32)
        return jax.lax.dynamic_update_index_in_dim(full_cache, zeros, slot, batch_idx)
      elif path_key == "cache_prefill_segment_id":
        s = list(full_cache.shape)
        s[batch_idx] = 1
        zeros = jnp.zeros(tuple(s), dtype=jnp.int32)
        ## zero out in case prefill cache is too small to cover
        full_cache = jax.lax.dynamic_update_index_in_dim(full_cache, zeros, slot, batch_idx)
        ## copy prefill cachce
        full_cache = jax.lax.dynamic_update_index_in_dim(full_cache, partial_cache, slot, batch_idx)
        return full_cache
      elif path_key == "cached_ar_lengths":
        return full_cache.at[slot].set(0)
      elif path_key in [
          "cached_prefill_key",
          "cached_prefill_value",
          "cached_prefill_key_scale",
          "cached_prefill_value_scale",
      ]:
        return jax.lax.dynamic_update_index_in_dim(full_cache, partial_cache, slot, batch_idx)
      else:
        raise ValueError(f"We don't have a strategy for inserting {path_key}")

    if self.config.attention == "paged":

      def _copy_paged(path, prefix_cache, decode_state_cache):
        if path[-2].key == "page_manager":
          return prefix_cache
        path_key = path[-1].key
        if path_key in ["key_pages", "value_pages"]:

          def _update_pages(prefix_page_idx, state):
            decode_state_pages, prefix_pages, page_map = state
            prefix_page = jax.lax.dynamic_index_in_dim(prefix_pages, prefix_page_idx, axis=1)
            decode_state_pages = jax.lax.dynamic_update_slice_in_dim(
                decode_state_pages, prefix_page, page_map[prefix_page_idx], axis=1
            )
            return decode_state_pages, prefix_pages, page_map

          decode_state_cache, _, _ = jax.lax.fori_loop(
              0,
              prefix["cache"]["page_manager"]["num_pages_used"].value[slot],
              _update_pages,
              (decode_state_cache, prefix_cache, prefix["cache"]["page_manager"]["page_map"].value[slot]),
          )
          return decode_state_cache
        else:
          raise ValueError(f"We don't have a strategy for inserting {path_key} for paged attention.")

      inserted_cache = jax.tree_util.tree_map_with_path(
          _copy_paged,
          unboxed_prefix["cache"],
          decode_state["cache"],
      )
    else:
      inserted_cache = jax.tree_util.tree_map_with_path(
          copy,
          unboxed_prefix["cache"],
          decode_state["cache"],
          self.kv_cache_annotations_named,
      )
    inserted_logits = jax.lax.dynamic_update_index_in_dim(decode_state["logits"], unboxed_prefix["logits"], slot, 0)
    inserted_next_pos = jax.lax.dynamic_update_index_in_dim(decode_state["next_pos"], unboxed_prefix["next_pos"], slot, 0)
    inserted_generated_tokens = jax.lax.dynamic_update_index_in_dim(
        decode_state["generated_tokens"],
        unboxed_prefix["generated_tokens"],
        slot,
        0,
    )
    inserted_tokens = jax.lax.dynamic_update_index_in_dim(decode_state["tokens"], unboxed_prefix["tokens"], slot, 0)

    inserted_logits = jax.lax.with_sharding_constraint(inserted_logits, self.replicated_sharding)
    inserted_generated_tokens = jax.lax.with_sharding_constraint(inserted_generated_tokens, self.replicated_sharding)
    inserted_next_pos = jax.lax.with_sharding_constraint(inserted_next_pos, self.replicated_sharding)
    inserted_tokens = jax.lax.with_sharding_constraint(inserted_tokens, self.replicated_sharding)
    inserted_cache = jax.lax.with_sharding_constraint(inserted_cache, self.kv_cache_shardings)

    return {
        "logits": inserted_logits,
        "cache": inserted_cache,
        "next_pos": inserted_next_pos,
        "generated_tokens": inserted_generated_tokens,
        "tokens": inserted_tokens,
    }

  @functools.partial(
      jax.jit,
      static_argnums=(0,),
      static_argnames=(
          "num_prompts",
          "seq_len",
      ),
      donate_argnums=(
          1,
          2,
      ),
  )
  def insert_partial(
      self,
      prefix: PackedPrefix,
      decode_state: DecodeState,
      cache: Any,
      slots: jax.Array,
      *,
      start_indices: jax.Array,
      num_prompts: int,
      seq_len: int,
  ) -> DecodeState:
    """Insert into KV cache"""
    unboxed_prefix = max_utils.unbox_logicallypartioned(prefix)
    cache_unboxed = max_utils.unbox_logicallypartioned(cache)
    cache_unboxed = self._maybe_unstack_prefill_result_cache(cache_unboxed)
    start_idx = 0
    slot = slots[0]

    def copy(path, partial_cache, full_cache, annotations):
      path_key = path[-1].key
      if path_key in [
          "cache_ar_index",
          "cached_ar_key",
          "cached_ar_value",
          "cached_ar_key_scale",
          "cached_ar_value_scale",
      ]:
        return full_cache  # we don't even zero these out because we can mask them out.

      batch_idx = -1
      if "cache_batch" in annotations:
        batch_idx = annotations.index("cache_batch")
      elif "cache_scale_batch" in annotations:
        batch_idx = annotations.index("cache_scale_batch")

      if batch_idx < 0:
        raise ValueError(f"Batch index {batch_idx=} shouldn't be less than zero for {path_key}, got {annotations=}")

      if path_key == "cache_ar_segment_id":
        ### goal: zero this out in case there is existing data
        zeros = jnp.zeros((1, self.config.max_target_length - self.config.max_prefill_predict_length), dtype=jnp.int32)
        return jax.lax.dynamic_update_index_in_dim(full_cache, zeros, slot, batch_idx)
      elif path_key == "cache_prefill_segment_id":
        zeros = jnp.zeros((1, self.config.max_prefill_predict_length), dtype=jnp.int32)
        ## zero out in case prefill cache is too small to cover
        full_cache = jax.lax.dynamic_update_index_in_dim(full_cache, zeros, slot, batch_idx)
        # In case partial_cache is too small to slice at the given index, pad it with an extra seqlen
        if i == num_prompts - 1:
          pad = jnp.zeros((1, seq_len), dtype=int)
          partial_cache = jnp.concatenate([partial_cache, pad], axis=1)
        ## copy prefill cache
        partial_cache = jax.lax.dynamic_slice(partial_cache, (0, start_idx), (1, seq_len))
        partial_cache = (partial_cache == partial_cache[0, 0]).astype(int)
        full_cache = jax.lax.dynamic_update_index_in_dim(full_cache, partial_cache, slot, batch_idx)
        return full_cache
      elif path_key == "cached_ar_lengths":
        return full_cache.at[slot].set(0)
      elif path_key in [
          "cached_prefill_key",
          "cached_prefill_value",
          "cached_prefill_key_scale",
          "cached_prefill_value_scale",
      ]:
        seqlen_index = self.config.prefill_cache_axis_order.split(",").index("1")
        start_indices = [0, 0, 0, 0]
        start_indices[seqlen_index] = start_idx
        slice_size = list(partial_cache.shape)
        slice_size[seqlen_index] = seq_len

        slice_size = tuple(slice_size)
        # Same as in prefill_segment_id processing
        if i == num_prompts - 1:
          pad = jnp.zeros(slice_size, dtype=partial_cache.dtype)
          partial_cache = jnp.concatenate([partial_cache, pad], axis=seqlen_index)
        partial_cache = jax.lax.dynamic_slice(partial_cache, start_indices, slice_size)

        return jax.lax.dynamic_update_index_in_dim(full_cache, partial_cache, slot, batch_idx)
      else:
        raise ValueError(f"We don't have a strategy for inserting {path_key}")

    inserted_cache = decode_state["cache"]
    inserted_logits = decode_state["logits"]
    inserted_next_pos = decode_state["next_pos"]
    inserted_generated_tokens = decode_state["generated_tokens"]
    inserted_tokens = decode_state["tokens"]

    for i in range(num_prompts):
      start_idx = start_indices[i]
      slot = slots[i]
      inserted_cache = jax.tree_util.tree_map_with_path(copy, cache_unboxed, inserted_cache, self.kv_cache_annotations_named)
      inserted_logits = jax.lax.dynamic_update_index_in_dim(inserted_logits, unboxed_prefix["logits"][i, ...], slot, 0)
      inserted_next_pos = jax.lax.dynamic_update_index_in_dim(inserted_next_pos, unboxed_prefix["next_pos"][i, ...], slot, 0)
      inserted_generated_tokens = jax.lax.dynamic_update_index_in_dim(
          inserted_generated_tokens,
          unboxed_prefix["generated_tokens"][i, ...],
          slot,
          0,
      )
      inserted_tokens = jax.lax.dynamic_update_index_in_dim(inserted_tokens, unboxed_prefix["tokens"][i, ...], slot, 0)

    inserted_logits = jax.lax.with_sharding_constraint(inserted_logits, self.replicated_sharding)
    inserted_generated_tokens = jax.lax.with_sharding_constraint(inserted_generated_tokens, self.replicated_sharding)
    inserted_next_pos = jax.lax.with_sharding_constraint(inserted_next_pos, self.replicated_sharding)
    inserted_tokens = jax.lax.with_sharding_constraint(inserted_tokens, self.replicated_sharding)
    inserted_cache = jax.lax.with_sharding_constraint(inserted_cache, self.kv_cache_shardings)

    return {
        "logits": inserted_logits,
        "cache": inserted_cache,
        "next_pos": inserted_next_pos,
        "generated_tokens": inserted_generated_tokens,
        "tokens": inserted_tokens,
    }

  def get_prefix_destination_sharding(self) -> Any:
    return {
        "logits": self.replicated_sharding,
        "cache": self.prefill_kv_cache_shardings,
        "next_pos": self.replicated_sharding,
        "generated_tokens": self.replicated_sharding,
        "tokens": self.replicated_sharding,
    }

  def get_tokenizer(self) -> tokenizer_pb2.TokenizerParameters:
    """Return a protobuf of tokenizer info, callable from Py or C++."""
    return tokenizer_pb2.TokenizerParameters(path=self.config.tokenizer_path, extra_ids=0)

  def build_tokenizer(self, metadata: tokenizer_pb2.TokenizerParameters) -> tokenizer_api.Tokenizer:
    """Return a tokenizer"""
    if "tiktoken" in metadata.path:
      return token_utils.TikToken(metadata)
    else:
      return token_utils.SentencePieceTokenizer(metadata)

  def init_decode_state(
      self,
      *args,  # pylint: disable=unused-argument
      rng: Optional[PRNGKeyType] = None,
      **kwargs,  # pylint: disable=unused-argument
  ) -> DecodeState:
    """Initialises any state which a generation step transforms."""

    if rng is None:
      rng = jax.random.PRNGKey(0)

    # pylint: disable=unused-argument
    def init(abstract_params):
      x = jnp.ones(
          (int(self.config.per_device_batch_size * jax.device_count()), 1),
          dtype=jnp.int32,
      )
      _, cache = self.model.apply(
          abstract_params,
          x,
          x,
          enable_dropout=False,
          model_mode=common_types.MODEL_MODE_AUTOREGRESSIVE,
          rngs={"params": rng},
          mutable=["cache"],
      )

      next_pos = jnp.zeros(
          (int(self.config.per_device_batch_size * jax.device_count()), 1),
          dtype=jnp.int32,
      )
      generated_tokens = jnp.zeros(
          (int(self.config.per_device_batch_size * jax.device_count()), 1),
          dtype=jnp.int32,
      )
      tokens = jnp.zeros(
          (int(self.config.per_device_batch_size * jax.device_count()), 1),
          dtype=jnp.int32,
      )
      return {
          "logits": jnp.zeros(
              (
                  int(self.config.per_device_batch_size * jax.device_count()),
                  1,
                  self.config.vocab_size,
              )
          ),
          "cache": cache["cache"],
          "next_pos": next_pos,
          "generated_tokens": generated_tokens,
          "tokens": tokens,
      }

    with nn_partitioning.axis_rules(self.config.logical_axis_rules):
      abstract_outputs = jax.eval_shape(init, self.abstract_params)
    logical_annotations = nn.get_partition_spec(abstract_outputs)

    with self._mesh, nn_partitioning.axis_rules(self.config.logical_axis_rules):
      mesh_annotations = nn.logical_to_mesh(logical_annotations)

    shardings = jax.tree_util.tree_map(
        lambda mesh_annotation: jax.sharding.NamedSharding(self._mesh, mesh_annotation),
        mesh_annotations,
    )

    @functools.partial(jax.jit, out_shardings=shardings)
    def initialize():
      return jax.tree_util.tree_map(lambda x: jnp.zeros(x.shape, x.dtype), abstract_outputs)

    init_state = initialize()
    cache = init_state["cache"]

    def is_lp(k):
      return isinstance(k, flax.linen.spmd.LogicallyPartitioned)

    self.kv_cache_annotations_named = jax.tree_util.tree_map(lambda x: tuple(x.names), cache, is_leaf=is_lp)
    zeroed = max_utils.unbox_logicallypartioned(init_state)
    return zeroed

  @property
  def max_concurrent_decodes(self) -> int:
    """Free slots."""
    return int(self.config.per_device_batch_size * jax.device_count())

  @property
  def max_prefill_length(self) -> int:
    """Maximum prefill length."""
    return int(self.config.max_prefill_predict_length)

  @property
  def use_chunked_prefill(self) -> bool:
    """Maximum prefill length."""
    return self.config.use_chunked_prefill

  @property
  def prefill_chunk_size(self) -> int:
    """Maximum prefill length."""
    return int(self.config.prefill_chunk_size)

  @property
  def samples_per_slot(self) -> int:
    """Number of samples per slot."""
    return 1

  @property
  def mesh(self) -> jax.sharding.Mesh:
    return self._mesh

  @property
  def colocated_cpus(self) -> None:
    """CPU devices colocated with the engine's accelerators."""
    raise NotImplementedError


def set_engine_vars_from_base_engine(
    engine: MaxEngine,
    base_engine: MaxEngine,
    rng: PRNGKeyType,
):
  """Set internal vars from base_engine, which has already loaded the checkpoint and has sharding,
  mesh, and kv cache related vars set.
  """
  if base_engine.model.quant:
    engine.model.quant.quant_mode = base_engine.model.quant.quant_mode
  engine.state_mesh_annotations = base_engine.state_mesh_annotations
  engine.abstract_params = base_engine.abstract_params
  engine.kv_cache_annotations = max_utils.get_kv_cache_annotations(engine.model, engine.config, rng, engine.mesh)  # pylint: disable=protected-access
  engine.kv_cache_shardings = jax.tree_util.tree_map(
      lambda x: jax.sharding.NamedSharding(engine.mesh, x),
      engine.kv_cache_annotations,  # pylint: disable=protected-access
  )


def create_engine_from_config_flags(batch_size, max_prefill_predict_length, max_target_length, args_str):
  """Create new MaxEngine instance with given batch_size, prefill and target lengths, and any config
  params provided through `args_str`.
  """
  args = {}
  args["scan_layers"] = "false"
  args["async_checkpointing"] = "false"
  args["ici_fsdp_parallelism"] = "1"
  args["ici_autoregressive_parallelism"] = "1"
  args["ici_tensor_parallelism"] = "-1"
  args["weight_dtype"] = "bfloat16"
  args["attention"] = "dot_product"

  # batch and cache related
  args["max_prefill_predict_length"] = f"{max_prefill_predict_length}"
  args["max_target_length"] = f"{max_target_length}"
  args["per_device_batch_size"] = f"{batch_size}"
  print(f"Command line args: {args_str}")
  cmd_args = args_str.split(" ")
  for cmd_arg in cmd_args:
    k, v = cmd_arg.split("=")
    args[k.strip()] = v.strip()
  assert "load_parameters_path" in args, "load_parameters_path must be defined"
  updated_args = ["MaxText/maxengine_server.py", "../configs/base.yml"]
  for k, v in args.items():
    option = f"{k}={v}"
    updated_args.append(option)
  print(f"Invoking maxengine with args:\n \t{updated_args}")
  cfg = pyconfig.initialize(updated_args)
  engine = MaxEngine(cfg)
  return engine



# Copyright 2024 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Contains common functions for configuring Jetstream server metrics"""

import os
import re
from typing import Optional
import shortuuid
from prometheus_client import Counter, Gauge, Histogram
from jetstream.engine.token_utils import DEFAULT_PREFILL_BUCKETS


class JetstreamMetricsCollector:
  """Wrapper class should be used to assure all metrics have proper tags"""

  _initialized: bool = False
  _model_name: str
  universal_labels = {"id": os.getenv("HOSTNAME", shortuuid.uuid())}

  def __new__(cls, model_name: Optional[str] = None):
    if not hasattr(cls, "instance"):
      cls.instance = super(JetstreamMetricsCollector, cls).__new__(cls)
    return cls.instance

  def __init__(self, model_name: Optional[str] = None):
    if hasattr(self, "_initialized") and self._initialized:
      return
    self._initialized = True

    # '-'s are common in model names but invalid in prometheus labels
    # these are replaced with '_'s
    if model_name is not None:
      sanitized_model_name = model_name.replace("-", "_")
      if sanitized_model_name == "":
        print("No model name provided, omitting from metrics labels")
      elif not bool(
          re.match(r"^[a-zA-Z_:][a-zA-Z0-9_:]*$", sanitized_model_name)
      ):
        print(
            "Provided model name cannot be used to label prometheus metrics",
            "(does not match ^[a-zA-Z_:][a-zA-Z0-9_:]*$)",
            "omitting from metrics labels",
        )
      else:
        self.universal_labels["model_name"] = sanitized_model_name
    universal_label_names = list(self.universal_labels.keys())

    # Metric definitions
    self._prefill_backlog = Gauge(
        name="jetstream_prefill_backlog_size",
        documentation="Size of prefill queue",
        labelnames=universal_label_names,
    )

    self._transfer_backlog = Gauge(
        name="jetstream_transfer_backlog_size",
        documentation="Size of transfer queue",
        labelnames=universal_label_names + ["idx"],
    )

    self._generate_backlog = Gauge(
        name="jetstream_generate_backlog_size",
        documentation="Size of generate queue",
        labelnames=universal_label_names + ["idx"],
    )

    self._queue_duration = Histogram(
        name="jetstream_queue_duration",
        documentation="The total time each request spends enqueued in seconds",
        labelnames=universal_label_names,
        buckets=[
            0.01,
            0.02,
            0.05,
            0.1,
            0.2,
            0.5,
            1.0,
            2.0,
            5.0,
            10.0,
            20.0,
            50.0,
            100.0,
        ],
    )

    self._slots_used_percentage = Gauge(
        name="jetstream_slots_used_percentage",
        documentation="The percentage of decode slots currently being used",
        labelnames=universal_label_names + ["idx"],
    )
    self._model_load_time = Gauge(
        name="jetstream_model_load_time",
        documentation="Total time taken to load the model",
        labelnames=universal_label_names,
    )
    self._server_startup_latency = Gauge(
        name="jetstream_server_startup_latency",
        documentation="Total time taken to start the Jetstream server",
        labelnames=universal_label_names,
    )
    self._request_input_length = Histogram(
        name="jetstream_request_input_length",
        documentation="Number of input tokens per request",
        labelnames=universal_label_names,
        buckets=DEFAULT_PREFILL_BUCKETS,
    )
    self._request_output_length = Histogram(
        name="jetstream_request_output_length",
        documentation="Number of output tokens per request",
        labelnames=universal_label_names,
        buckets=[
            1,
            2,
            5,
            10,
            20,
            50,
            100,
            200,
            500,
            1000,
            2000,
            5000,
            10000,
            20000,
            50000,
            100000,
            200000,
            500000,
            1000000,
            2000000,
        ],
    )
    self._request_success_count = Counter(
        name="jetstream_request_success_count",
        documentation="Number of requests successfully completed",
        labelnames=universal_label_names,
    )

    self._time_to_first_token = Histogram(
        name="jetstream_time_to_first_token",
        documentation="Time to first token per request in seconds",
        labelnames=universal_label_names,
        buckets=[
            0.001,
            0.005,
            0.01,
            0.02,
            0.04,
            0.06,
            0.08,
            0.1,
            0.25,
            0.5,
            0.75,
            1.0,
            2.5,
            5.0,
            7.5,
            10.0,
        ],
    )

    self._time_per_output_token = Histogram(
        name="jetstream_time_per_output_token",
        documentation="Average time per output token per request in seconds",
        labelnames=universal_label_names,
        buckets=[
            0.01,
            0.025,
            0.05,
            0.075,
            0.1,
            0.15,
            0.2,
            0.3,
            0.4,
            0.5,
            0.75,
            1.0,
            2.5,
        ],
    )

    self._time_per_prefill_token = Histogram(
        name="jetstream_time_per_prefill_token",
        documentation="Prefill time per token per request in seconds",
        labelnames=universal_label_names,
        buckets=[
            0.00001,
            0.00002,
            0.00005,
            0.0001,
            0.0002,
            0.0005,
            0.001,
            0.002,
            0.005,
            0.01,
            0.02,
            0.05,
            0.1,
        ],
    )

    self._time_per_request = Histogram(
        name="jetstream_time_per_request",
        documentation="End to end request latency in seconds",
        labelnames=universal_label_names,
        buckets=[1.0, 2.5, 5.0, 10.0, 15.0, 20.0, 30.0, 40.0, 50.0, 60.0],
    )

    self._wait_time_per_request = Histogram(
        name="jetstream_wait_time_per_request",
        documentation="Time each request is not being prefilled or decoded",
        labelnames=universal_label_names,
        buckets=[
            0.01,
            0.02,
            0.05,
            0.1,
            0.2,
            0.5,
            1.0,
            2.0,
            5.0,
            10.0,
            20.0,
            50.0,
            100.0,
        ],
    )

  def get_prefill_backlog_metric(self):
    return self._prefill_backlog.labels(**self.universal_labels)

  def get_transfer_backlog_metric(self, idx: int):
    return self._transfer_backlog.labels(**self.universal_labels, idx=idx)

  def get_generate_backlog_metric(self, idx: int):
    return self._generate_backlog.labels(**self.universal_labels, idx=idx)

  def get_queue_duration(self):
    return self._queue_duration.labels(**self.universal_labels)

  def get_slots_used_percentage_metric(self, idx: int):
    return self._slots_used_percentage.labels(**self.universal_labels, idx=idx)

  def get_server_startup_latency_metric(self):
    return self._server_startup_latency.labels(**self.universal_labels)

  def get_model_load_time_metric(self):
    return self._model_load_time.labels(**self.universal_labels)

  def get_time_to_first_token(self):
    return self._time_to_first_token.labels(**self.universal_labels)

  def get_time_per_output_token(self):
    return self._time_per_output_token.labels(**self.universal_labels)

  def get_time_per_prefill_token(self):
    return self._time_per_prefill_token.labels(**self.universal_labels)

  def get_time_per_request(self):
    return self._time_per_request.labels(**self.universal_labels)

  def get_wait_time_per_request(self):
    return self._wait_time_per_request.labels(**self.universal_labels)

  def get_request_input_length(self):
    return self._request_input_length.labels(**self.universal_labels)

  def get_request_output_length(self):
    return self._request_output_length.labels(**self.universal_labels)

  def get_request_success_count_metric(self):
    return self._request_success_count.labels(**self.universal_labels)



# Copyright 2024 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Tests gRPC server end-to-end.

See orchestrator test for why these characters specifically will be the
response.
"""

import unittest
from typing import Any, Type

import grpc
import portpicker
import requests
from parameterized import parameterized

from jetstream.core import config_lib
from jetstream.core import server_lib
from jetstream.core.proto import jetstream_pb2
from jetstream.core.proto import jetstream_pb2_grpc
from jetstream.engine import engine_api


class ServerTest(unittest.TestCase):

  @parameterized.expand(
      [
          # Uses weight 2 for prefill, 4 for decode.
          (
              config_lib.CPUTestServer,
              True,
              ["Ċ", "Ō", "Ɵ", ""],
              [266, 332, 415, None],
              [None, None],
          ),
          # Uses the same prefill / generate weights (2).
          (
              config_lib.InterleavedCPUTestServer,
              True,
              ["Ċ", "Ə", "ɖ", ""],
              [266, 399, 598, None],
              [None],
          ),
          # Disable the metrics server.
          (
              config_lib.InterleavedCPUTestServer,
              False,
              ["Ċ", "Ə", "ɖ", ""],
              [266, 399, 598, None],
              [None],
          ),
      ]
  )
  def test_server(
      self,
      config: Type[config_lib.ServerConfig],
      metrics_enabled: bool,
      expected_text: list[str],
      expected_token_ids: list[int | None],
      devices: list[Any],
  ):
    """Sets up a server and requests token responses."""
    ######################### Server side ######################################
    port = portpicker.pick_unused_port()
    metrics_port = portpicker.pick_unused_port()

    print("port: " + str(port))
    credentials = grpc.local_server_credentials()

    server = server_lib.run(
        port=port,
        config=config,
        devices=devices,
        credentials=credentials,
        metrics_server_config=config_lib.MetricsServerConfig(
            port=metrics_port, model_name="some_model_name"
        )
        if metrics_enabled is True
        else None,
    )
    ###################### Requester side ######################################

    # if prometheus not configured, assert no metrics collector on Driver
    if metrics_enabled is not True:
      self.assertIsNone(server._driver._metrics_collector)  # pylint: disable=protected-access

    with grpc.secure_channel(
        f"localhost:{port}", grpc.local_channel_credentials()
    ) as channel:
      stub = jetstream_pb2_grpc.OrchestratorStub(channel)

      healthcheck_request = jetstream_pb2.HealthCheckRequest()
      healthcheck_response = stub.HealthCheck(healthcheck_request)
      self.assertTrue(healthcheck_response.is_live)

      # The string representation of np.array([[65, 66]]), [2] will be prepended
      # as BOS
      text = "AB"
      request = jetstream_pb2.DecodeRequest(
          text_content=jetstream_pb2.DecodeRequest.TextContent(text=text),
          max_tokens=3,
      )
      iterator = stub.Decode(request)
      counter = 0
      for resp in iterator:
        output_text = resp.stream_content.samples[0].text
        token_ids = resp.stream_content.samples[0].token_ids
        output_token_id = token_ids[0] if len(token_ids) > 0 else None
        print(f"actual output: {output_text=} {output_token_id=}")
        self.assertEqual(output_text, expected_text[counter])
        self.assertEqual(output_token_id, expected_token_ids[counter])
        counter += 1
      # assert prometheus server is running and responding
      if metrics_enabled is True:
        self.assertIsNotNone(server._driver._metrics_collector)  # pylint: disable=protected-access
        self.assertEqual(
            requests.get(
                f"http://localhost:{metrics_port}", timeout=5
            ).status_code,
            requests.status_codes.codes["ok"],
        )
      server.stop()
      pass

  def test_jax_profiler_server(self):
    port = portpicker.pick_unused_port()
    print("port: " + str(port))
    credentials = grpc.local_server_credentials()
    # Now test server with prometheus config
    server = server_lib.run(
        port=port,
        config=config_lib.InterleavedCPUTestServer,
        devices=[None],
        credentials=credentials,
        enable_jax_profiler=True,
    )
    self.assertIsNotNone(server)
    server.stop()

  def test_get_devices(self):
    self.assertEqual(len(server_lib.get_devices()), 1)

  def test_model_warmup(self):
    port = portpicker.pick_unused_port()

    print("port: " + str(port))
    credentials = grpc.local_server_credentials()

    server = server_lib.run(
        port=port,
        config=config_lib.InterleavedCPUTestServer,
        devices=[None],
        credentials=credentials,
        enable_model_warmup=True,
    )
    # await _wait_for_server_ready(port, timeout_sec=5)
    with grpc.secure_channel(
        f"localhost:{port}", grpc.local_channel_credentials()
    ) as channel:
      stub = jetstream_pb2_grpc.OrchestratorStub(channel)

      healthcheck_request = jetstream_pb2.HealthCheckRequest()
      healthcheck_response = stub.HealthCheck(healthcheck_request)
      self.assertTrue(healthcheck_response.is_live)

      for pe in server._driver._prefill_engines:  # pylint: disable=protected-access
        self.assertIsInstance(pe, engine_api.JetStreamEngine)
        self.assertTrue(pe.warm)

      for ge in server._driver._generate_engines:  # pylint: disable=protected-access
        self.assertIsInstance(ge, engine_api.JetStreamEngine)
        self.assertTrue(ge.warm)

    server.stop()



"""
Copyright 2025 Google LLC

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

     https://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""

import glob
import os
import torch
from safetensors import safe_open
import pathlib
import jax
import numpy as np
import argparse


def load_hf(hf_checkpoint_folder):
  safetensor_files = glob.glob(os.path.join(hf_checkpoint_folder, "*.safetensors"))

  hf_tensor = {}
  for st_f in safetensor_files:
    with safe_open(st_f, framework="pt", device="cpu") as f:
      for key in f.keys():
        hf_tensor[key] = f.get_tensor(key).to(torch.float16)
  return hf_tensor


def load_meta(meta_checkpoint_folder):
  meta_tensor = {}
  ckpt_paths = sorted(pathlib.Path(meta_checkpoint_folder).glob("[!.]*.pth"))
  for ckpt_path in ckpt_paths:
    meta_tensor = torch.load(ckpt_path, map_location="cpu")
  return meta_tensor


def compare_pytrees(tree1, tree2, atol=0.001):
  """
  Compares two JAX pytrees to check if all leaf values are within the given absolute tolerance.

  Args:
      tree1: First pytree.
      tree2: Second pytree.
      atol: Absolute tolerance for comparison (default: 0.001).

  Returns:
      A boolean indicating if all leaf values are within the specified range.
  """
  # Ensure both trees have the same structure
  if jax.tree_util.tree_structure(tree1) != jax.tree_util.tree_structure(tree2):
    print(
        "Pytrees have different structures! Tree1:"
        f"{jax.tree_util.tree_structure(tree1)} \n\n\n"
        f"Tree2: {jax.tree_util.tree_structure(tree2)}"
    )
    return

  # Compare leaves with names
  def get_named_leaves(pytree, parent_key=""):
    named_leaves = {}
    for key, value in pytree.items():
      new_key = f"{parent_key}.{key}" if parent_key else key
      if isinstance(value, dict):
        named_leaves.update(get_named_leaves(value, new_key))
      else:
        named_leaves[new_key] = value
    return named_leaves

  named_leaves1 = get_named_leaves(tree1)
  named_leaves2 = get_named_leaves(tree2)

  print(f"There are {len(named_leaves1.keys())} leaves to check.")
  for key in named_leaves1:  # pylint: disable=C0206
    if key not in named_leaves2:
      print(f"Missing key in second tree: {key}")
      return
    try:
      if not np.allclose(named_leaves1[key], named_leaves2[key], atol=atol):
        print(f"Mismatch at leaf '{key}' with shape {named_leaves1[key].shape}:\n")
        for i in range(10):
          print(f"{named_leaves1[key][..., i, :]}\n")
        print("The second tensor:\n")
        for i in range(10):
          print(f"{named_leaves2[key][..., i, :]}\n")
        return
    except:  # pylint: disable=W0702
      print(f"The issue is with {key}")

  print(f"All {len(named_leaves1.keys())} leaves match within tolerance.")


def main():
  parser = argparse.ArgumentParser(description="Compares the original checkpoint and converted back checkpoint.")
  parser.add_argument(
      "--original_ckpt",
      type=str,
      default="",
      help="The original huggingface checkpoint",
  )
  parser.add_argument(
      "--converted_ckpt",
      type=str,
      default="",
      help="The original huggingface checkpoint",
  )
  args = parser.parse_args()

  hf_tensor = load_hf(args.original_ckpt)
  meta_tensor = load_hf(args.converted_ckpt)

  compare_pytrees(hf_tensor, meta_tensor)


if __name__ == "__main__":
  main()



"""
Copyright 2024 Google LLC

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

     https://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""

"""Chunked prefill TPU kernel with paged kv cache."""

import jax
from jax import numpy as jnp
from jax.experimental import pallas as pl
from jax.experimental.pallas import tpu as pltpu
from inference.kernel.attention.tpu.paged_attention import *


DEFAULT_MASK_VALUE = -2.3819763e38  # Set to a large negative number.


def chunked_prefill_attention_impl(
    length_ref,  # shape: (1,), smem,
    page_indices_ref,  # shape: (max_seq_len // page_size), smem,
    buffer_index_ref,  # shape: (1,), smem,
    q_ref,  # shape: (group_size, chunk, head_dim), vmem,
    k_pages_hbm_ref,  # shape: (num_kv_heads, num_pages, page_size, head_dim), hbm
    v_pages_hbm_ref,  # shape: (num_kv_heads, num_pages, page_size, head_dim), hbm
    out_ref,  # shape: (group_size, chunk, head_dim), vmem,
    l_ref,  # shape: (group_size, chunk, 1), vmem,
    m_ref,  # shape: (group_size, chunk, 1), vmem,
    k_vmem_buffer,  # shape: (2, page_per_chunk, page_size, head_dim), vmem,
    v_vmem_buffer,  # shape: (2, page_per_chunk, page_size, head_dim), vmem,
    sem,
):
  h = pl.program_id(0)
  page_size = k_pages_hbm_ref.shape[2]
  head_dim = k_pages_hbm_ref.shape[3]
  group_size = q_ref.shape[0]
  num_kv_heads = k_pages_hbm_ref.shape[0]
  chunk_size = q_ref.shape[1]
  length = length_ref[0]
  q_chunk_idx = jax.lax.div(length, chunk_size)
  reminder = jax.lax.rem(length, chunk_size)
  q_chunk_idx -= jnp.where(reminder > 0, 0, 1)
  out_ref[...] = jnp.zeros_like(out_ref)

  def create_kv_async_copy_descriptors(h, i, buffer_index):
    pages_to_load = chunk_size // page_size
    page_offset = i * pages_to_load
    async_copy_k = MultiPageAsyncCopyDescriptor(
        k_pages_hbm_ref,
        None,
        k_vmem_buffer.at[buffer_index],
        None,
        sem,
        page_indices_ref,
        page_offset,
        pages_to_load,
        head_index=h,
    )
    async_copy_v = MultiPageAsyncCopyDescriptor(
        v_pages_hbm_ref,
        None,
        v_vmem_buffer.at[buffer_index],
        None,
        sem,
        page_indices_ref,
        page_offset,
        pages_to_load,
        head_index=h,
    )
    return async_copy_k, async_copy_v

  def next_block_indice(h, i):
    return jax.lax.cond(
        (i + 1) * chunk_size < length, lambda: (h, i + 1), lambda: (h + 1, 0)
    )

  def per_kv_chunk_body(i, _):
    @pl.when((i * chunk_size) < length)
    def body():
      buffer_index = buffer_index_ref[0]

      @pl.when(i == 0)
      def init():
        m_ref[...] = jnp.full_like(m_ref, -jnp.inf)
        l_ref[...] = jnp.zeros_like(l_ref)

        @pl.when(h == 0)
        def prefetch_first_kv():
          # prefetch the first kv chunk.
          async_copy_k, async_copy_v = create_kv_async_copy_descriptors(
              h, i, buffer_index
          )
          async_copy_k.start()
          async_copy_v.start()

      next_h, next_i = next_block_indice(h, i)

      @pl.when((next_h < num_kv_heads) & (next_i <= q_chunk_idx))
      def prefetch_next_block():
        # prefetch the kv chunk for next iteration.
        next_buffer_index = jnp.where(buffer_index == 0, 1, 0)
        async_copy_next_k, async_copy_next_v = create_kv_async_copy_descriptors(
            next_h, next_i, next_buffer_index
        )

        async_copy_next_k.start()
        async_copy_next_v.start()
        buffer_index_ref[0] = next_buffer_index

      async_copy_k, async_copy_v = create_kv_async_copy_descriptors(
          h, i, buffer_index
      )

      k = async_copy_k.wait_and_get_loaded()
      v = async_copy_v.wait_and_get_loaded()

      mask_shape = (chunk_size, chunk_size)
      row_ids = jax.lax.broadcasted_iota(jnp.int32, mask_shape, 0)
      row_ids += q_chunk_idx * chunk_size
      col_ids = jax.lax.broadcasted_iota(jnp.int32, mask_shape, 1)
      col_ids += i * chunk_size
      causal_mask = col_ids <= row_ids
      causal_mask_value = jnp.where(causal_mask, 0.0, DEFAULT_MASK_VALUE)

      def per_group_body(group_idx, _):
        q = q_ref[group_idx]
        s = (
            jnp.einsum("td,sd->ts", q, k, preferred_element_type=jnp.float32)
            + causal_mask_value
        )
        # mask.
        s_max = jnp.max(s, axis=1, keepdims=True)

        prev_m = m_ref[group_idx]
        prev_l = l_ref[group_idx]

        cur_m = jnp.maximum(prev_m, s_max)
        cur_m_to_attn_size = jax.lax.broadcast_in_dim(
            cur_m, (chunk_size, chunk_size), (0, 1)
        )

        p = jnp.exp(s - cur_m_to_attn_size)

        cur_l = jnp.exp(prev_m - cur_m) * prev_l + jnp.sum(
            p, axis=1, keepdims=True
        )

        out = out_ref[group_idx]

        out_ref[group_idx, :, :] = (
            out
            * jax.lax.broadcast_in_dim(
                jnp.exp(prev_m - cur_m), (chunk_size, head_dim), (0, 1)
            )
            + p @ v
        ).astype(
            out_ref.dtype
        )  # p @ v  "ts,sd->td"

        m_ref[group_idx, :, :] = cur_m
        l_ref[group_idx, :, :] = cur_l
        return ()

      jax.lax.fori_loop(0, group_size, per_group_body, ())

    @pl.when(((i + 1) * chunk_size) >= length)
    def rescale():
      l = jax.lax.broadcast_in_dim(
          l_ref[...], (group_size, chunk_size, head_dim), (0, 1, 2)
      )
      out_ref[...] = (out_ref[...] / l).astype(out_ref.dtype)

    return ()

  # loop over k, v cache chunk.
  jax.lax.fori_loop(
      0, lax.div(length + chunk_size - 1, chunk_size), per_kv_chunk_body, ()
  )


# TODO: Change to firstly attend to the current chunk kv
# and then write to the KV Cache storage to avoid redundant
# KV Cache reading.
def chunked_prefill_attention(
    q: jax.Array,
    k_pages: jax.Array,
    v_pages: jax.Array,
    length: jax.Array,
    page_indices: jax.Array,
):
  """TPU chunked prefill attention."""
  chunk_size, num_attn_heads, head_dim = q.shape
  num_kv_heads, _, page_size, _ = k_pages.shape

  assert num_attn_heads % num_kv_heads == 0
  assert chunk_size % page_size == 0
  attn_group_size = num_attn_heads // num_kv_heads
  page_per_chunk = chunk_size // page_size

  # q shape as (num_attn_heads, chunk_size, head_dim)
  q = q.transpose((1, 0, 2))
  q = q / jnp.sqrt(head_dim)

  q_block_spec = pl.BlockSpec(
      (attn_group_size, chunk_size, head_dim), lambda i, *_: (i, 0, 0)
  )
  lm_block_spec = pl.BlockSpec(
      (attn_group_size, chunk_size, 1), lambda *_: (0, 0, 0)
  )
  lm_shape = jax.ShapeDtypeStruct(
      shape=(attn_group_size, chunk_size, 1), dtype=jnp.float32
  )
  # loop over Q chunk and num kv heads dimension.
  out, _, _ = pl.pallas_call(
      chunked_prefill_attention_impl,
      grid_spec=pltpu.PrefetchScalarGridSpec(
          num_scalar_prefetch=3,
          in_specs=[
              q_block_spec,
              pl.BlockSpec(memory_space=pltpu.TPUMemorySpace.ANY),
              pl.BlockSpec(memory_space=pltpu.TPUMemorySpace.ANY),
          ],
          out_specs=[
              q_block_spec,
              lm_block_spec,
              lm_block_spec,
          ],
          scratch_shapes=[
              pltpu.VMEM(
                  (
                      2,  # For double buffering during DMA copies.
                      page_per_chunk,
                      page_size,
                      head_dim,
                  ),
                  k_pages.dtype,
              ),  # k_pages buffer
              pltpu.VMEM(
                  (
                      2,  # For double buffering during DMA copies.
                      page_per_chunk,
                      page_size,
                      head_dim,
                  ),
                  v_pages.dtype,
              ),  # v_pages buffer
              pltpu.SemaphoreType.DMA,
          ],
          grid=(num_kv_heads,),
      ),
      out_shape=[
          jax.ShapeDtypeStruct(q.shape, q.dtype),
          lm_shape,
          lm_shape,
      ],
      # interpret=True,
      # debug=True
  )(
      jnp.reshape(length, (1,)),
      page_indices,
      jnp.asarray([0], jnp.int32),
      q,
      k_pages,
      v_pages,
  )
  out = out.transpose((1, 0, 2)).reshape(chunk_size, -1).astype(q.dtype)

  return out



# Copyright 2024 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from kithara.model.maxtext.maxtext.MaxText.maxengine import MaxEngine
import jax
import max_utils

class MaxtextInferenceEngine(MaxEngine):
    """This is a patched version of MaxEngine
    
    Changes: 
      -  Added the `load_existing_params` function to allow 
      the engine to run with an exisitng model instead of initiating
      a new one. 
    """
    def load_existing_params(self, params, rng=None) -> "Params":

        if rng is None:
            rng = jax.random.PRNGKey(0)

        self.abstract_params = jax.tree_util.tree_map(
            lambda x: (
                jax.ShapeDtypeStruct(shape=x.shape, dtype=x.dtype, sharding=x.sharding)
                if isinstance(x, jax.Array)
                else None
            ),
            params,
        )

        self.prefill_kv_cache_annotations = max_utils.get_prefill_kv_cache_annotations(
            self.model, self.config, rng, self._mesh
        )
        self.prefill_kv_cache_shardings = jax.tree_util.tree_map(
            lambda x: jax.sharding.NamedSharding(self._mesh, x),
            self.prefill_kv_cache_annotations,
        )

        if self.config.stack_prefill_result_cache:
            # Add extra axis for the axis generated by the stack.
            self.prefill_kv_cache_shardings = jax.tree_util.tree_map(
                lambda x: jax.sharding.NamedSharding(
                    self._mesh, jax.sharding.PartitionSpec(None, *x.spec)
                ),
                self.prefill_kv_cache_shardings,
            )
            self.prefill_kv_cache_shardings = self.prefill_kv_cache_shardings[
                "decoder"
            ]["layers_0"]

        self.kv_cache_annotations = max_utils.get_kv_cache_annotations(
            self.model, self.config, rng, self._mesh
        )
        self.kv_cache_shardings = jax.tree_util.tree_map(
            lambda x: jax.sharding.NamedSharding(self._mesh, x),
            self.kv_cache_annotations,
        )
        return params



"""
Copyright 2025 Google LLC

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

     https://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""

""" Tests for DeepSeek """
import math
import torch
import jax
import unittest
import jax.numpy as jnp
from layers import embeddings
import numpy as np


"""  
DeepSeek v3 PyTorch implementation of yarn rotary positional embedding.
See https://github.com/deepseek-ai/DeepSeek-V3/blob/2f7b80eecebf3d1c84da5a0d465f6639ea175012/inference/model.py#L294 for details.

"""


def precompute_freqs_cis(dim, seqlen, beta_fast=32, beta_slow=1, base=10000.0, factor=40) -> torch.Tensor:
  """
  Precomputes frequency-based complex exponential values for rotary positional embeddings.

  Args:
      args (ModelArgs): Model arguments containing positional embedding parameters.

  Returns:
      torch.Tensor: Precomputed complex exponential values for positional embeddings.
  """
  original_seq_len = 4096

  def find_correction_dim(num_rotations, dim, base, max_seq_len):
    """
    Computes the correction dimension for a given number of rotations in the rotary positional embedding.

    Args:
        num_rotations (float): Number of rotations to compute the correction for.
        dim (int): Dimensionality of the embedding space.
        base (float): Base value for the exponential computation.
        max_seq_len (int): Maximum sequence length.

    Returns:
        float: The correction dimension based on the input parameters.
    """
    return dim * math.log(max_seq_len / (num_rotations * 2 * math.pi)) / (2 * math.log(base))

  def find_correction_range(low_rot, high_rot, dim, base, max_seq_len):
    """
    Computes the range of correction dimensions for rotary positional embeddings.

    Args:
        low_rot (float): Lower bound for the number of rotations.
        high_rot (float): Upper bound for the number of rotations.
        dim (int): Dimensionality of the embedding space.
        base (float): Base value for the exponential computation.
        max_seq_len (int): Maximum sequence length.

    Returns:
        Tuple[int, int]: The range of correction dimensions (low, high), clamped to valid indices.
    """
    low = math.floor(find_correction_dim(low_rot, dim, base, max_seq_len))
    high = math.ceil(find_correction_dim(high_rot, dim, base, max_seq_len))
    return max(low, 0), min(high, dim - 1)

  def linear_ramp_factor(min, max, dim):
    """
    Computes a linear ramp function used to smooth values between a minimum and maximum range.

    Args:
        min (float): Minimum value for the ramp function.
        max (float): Maximum value for the ramp function.
        dim (int): Dimensionality of the ramp tensor.

    Returns:
        torch.Tensor: A tensor of shape (dim,) with values linearly interpolated between 0 and 1,
            clamped to the range [0, 1].
    """
    if min == max:
      max += 0.001
    linear_func = (torch.arange(dim, dtype=torch.float32) - min) / (max - min)
    ramp_func = torch.clamp(linear_func, 0, 1)
    return ramp_func

  freqs = 1.0 / (base ** (torch.arange(0, dim, 2, dtype=torch.float32) / dim))
  if seqlen > original_seq_len:
    low, high = find_correction_range(beta_fast, beta_slow, dim, base, original_seq_len)
    smooth = 1 - linear_ramp_factor(low, high, dim // 2)
    freqs = freqs / factor * (1 - smooth) + freqs * smooth

  t = torch.arange(seqlen)
  freqs = torch.outer(t, freqs)
  freqs_cis = torch.polar(torch.ones_like(freqs), freqs)
  return freqs_cis


def apply_rotary_emb(x: torch.Tensor, freqs_cis: torch.Tensor) -> torch.Tensor:
  """
  Applies rotary positional embeddings to the input tensor.

  Args:
      x (torch.Tensor): Input tensor with positional embeddings to be applied.
      freqs_cis (torch.Tensor): Precomputed complex exponential values for positional embeddings.

  Returns:
      torch.Tensor: Tensor with rotary embeddings applied.
  """
  dtype = x.dtype
  x = torch.view_as_complex(x.float().view(*x.shape[:-1], -1, 2))
  freqs_cis = freqs_cis.view(1, x.size(1), 1, x.size(-1))
  y = torch.view_as_real(x * freqs_cis).flatten(3)
  return y.to(dtype)


def to_jax(pt_tensor: torch.Tensor) -> jax.Array:
  return jnp.asarray(pt_tensor.detach().numpy())


class YarnRoPETest(unittest.TestCase):
  """Test for the Yarn RoPE implementation."""

  def test_rope_multiple_seq(self):
    hidden_dim = 32
    batch_size = 16
    num_heads = 8

    # Try multiple sequence lengths
    for seqlen in [128, 5000, 10000]:
      with self.subTest(seqlen=seqlen):
        # Precompute the frequencies for RoPE
        freqs_cis = precompute_freqs_cis(hidden_dim, seqlen=seqlen)

        # Create random queries and keys
        q = torch.randn(batch_size, seqlen, num_heads, hidden_dim)
        k = torch.randn(batch_size, seqlen, num_heads, hidden_dim)

        # Generate position ids
        position_ids = torch.arange(seqlen).unsqueeze(0).repeat(batch_size, 1)

        # Apply RoPE using the PyTorch implementation
        q_rope_pt = apply_rotary_emb(q, freqs_cis)
        k_rope_pt = apply_rotary_emb(k, freqs_cis)

        # Create and initialize the JAX Yarn RoPE
        model_jax = embeddings.YarnRotaryEmbedding(hidden_dim, seqlen)
        params = model_jax.init(jax.random.PRNGKey(0), to_jax(k))

        # Apply the JAX RoPE
        q_rope_jax = model_jax.apply(params, to_jax(q), to_jax(position_ids))
        # Infer positions from k.
        k_rope_jax = model_jax.apply(params, to_jax(k))

        # Compare outputs from the PyTorch and JAX implementations
        np.testing.assert_allclose(to_jax(q_rope_pt), q_rope_jax, rtol=1e-3, atol=0.05)
        np.testing.assert_allclose(to_jax(k_rope_pt), k_rope_jax, rtol=1e-3, atol=0.05)


if __name__ == "__main__":
  unittest.main()



# Copyright 2024 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Unit test for config_lib.py."""

import unittest
from parameterized import parameterized
from jetstream.core import config_lib


class TestConfigLib(unittest.TestCase):

  @parameterized.expand([("tpu=8", 8), ("v5e-8", 8), ("v5e=4", 4), ("v4-8", 4)])
  def test_slice_to_num_chips(self, accelerator_slice, expected_num_devices):
    got = config_lib.slice_to_num_chips(accelerator_slice)
    self.assertEqual(got, expected_num_devices)

  def test_get_engines_invalid(self):
    with self.assertRaises(ValueError):
      config_lib.get_engines(config_lib.InterleavedCPUTestServer, [])



"""
Copyright 2024 Google LLC

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

     https://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""

from absl.testing import absltest
import os
import jax.experimental.shard_map
import numpy as np
import jax.experimental
import jax.experimental.mesh_utils
import jax
from jax import numpy as jnp
from jax.sharding import NamedSharding, PartitionSpec as P
from inference import parallel
from inference import nn


class NormTest(absltest.TestCase):

  def _create_device_mesh(self):
    devices = jax.devices()
    return parallel.create_device_mesh(
        devices=devices,
        shape=(len(devices), 1),
    )

  def test_rmsnorm_per_device_forward(self):
    mesh = self._create_device_mesh()
    hidden_state_size = 128
    eps = 1e-6
    rmsnorm_layer = nn.RMSNorm(
        dim=hidden_state_size,
        eps=eps,
        parallel_config=parallel.RMSNormParallelConfig(
            mesh=mesh,
            activation_sharded=False,
        ),
    )
    distributed_rmsnorm_layer = nn.RMSNorm(
        dim=hidden_state_size,
        eps=eps,
        parallel_config=parallel.RMSNormParallelConfig(
            mesh=mesh,
            activation_sharded=True,
        ),
    )

    key = jax.random.PRNGKey(0)
    input = jax.random.uniform(key, (96, hidden_state_size))
    weight = jax.random.uniform(key, (hidden_state_size,))
    sharded_weight = jnp.copy(weight)

    rmsnorm_layer.load_weights_dict({"weight": weight})
    dis_weight = distributed_rmsnorm_layer.load_weights_dict(
        {"weight": sharded_weight}
    )
    expect = rmsnorm_layer(input)

    sharded_input = jax.device_put(
        input, NamedSharding(mesh, P(None, parallel.tp_axis_names()))
    )

    def distributed_rms_ag(weight, input):
      output = distributed_rmsnorm_layer.jittable_call(weight, input)
      return parallel.ops.all_gather(output, 1, parallel.tp_axis_names())

    got = jax.experimental.shard_map.shard_map(
        distributed_rms_ag,
        mesh,
        in_specs=(
            P(parallel.tp_axis_names()),
            P(None, parallel.tp_axis_names()),
        ),
        out_specs=P(None, None),
        check_rep=False,
    )(dis_weight, sharded_input)

    np.testing.assert_allclose(got, expect, atol=1e-6, rtol=1e-7)


if __name__ == "__main__":
  absltest.main()



# Copyright 2024 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Utils for engine operation."""

from typing import List, Sequence
from flax import struct
import numpy as np
from seqio.vocabularies import Vocabulary


@struct.dataclass
class TestTokenizer:
  """Tokenizer used for testing purposes."""

  def IdToPiece(self, integer: int) -> str:  # pylint: disable=invalid-name
    """In the real version, unlike encode_tf/decode_tf, doesn't strip trailing

    whitespace.
    """
    return chr(integer)

  def decode(self, tokens: np.ndarray):  # pylint: disable=invalid-name
    """Converts a numpy array into a string.

    Uses tokens[0] as we are doing streaming decode now
    """
    return chr(tokens[0])


@struct.dataclass
class TestVocab(Vocabulary):
  """Mock vocabulary used for tests.

  These methods are duplicative on the test vocab, but required to fit
  the seqio.Vocabulary interface.
  """

  pad_id = 0
  eos_id = 1
  bos_id = 2
  unk_id = 3
  stop_tokens = {pad_id, eos_id}
  _base_vocab_size = 2**16
  tokenizer: TestTokenizer = TestTokenizer()

  def _encode(self, s: str) -> Sequence[int]:
    """Converts a string into a integer sequenc."""
    # 'We use array methods, not python iterables so we don't
    # implement this method in the mock vocab.
    raise NotImplementedError

  def _decode(self, ids: np.ndarray):
    """Converts a numpy array into a string."""
    return "".join([chr(r) for r in list(ids) if r not in self.stop_tokens])

  def _encode_tf(self, s: str) -> np.ndarray:
    """Converts a string into a numpy array."""
    # We mock using numpy to avoid propagating tf dependencies.
    chars = np.array([ord(c) for c in s]).astype(np.int32)
    return chars

  def _decode_tf(self, ids: np.ndarray) -> List[str]:
    """Converts a numpy array into a string."""
    # We mock using numpy to avoid propagating tf dependencies.
    results = np.split(ids, ids.shape[0])
    return ["".join([chr(r) for r in list(line[0])]) for line in results]

  def decode(self, ids: np.ndarray, is_streaming=True):
    """Converts a numpy array into a string."""
    return is_streaming and self._decode(ids)

  def encode_tf(self, s: str) -> np.ndarray:
    """Converts a string into a numpy array."""
    return self._encode_tf(s)

  def decode_tf(self, ids: np.ndarray) -> List[str]:
    """Converts a numpy array into a string."""
    return self._decode_tf(ids)



# Copyright 2024 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Tests functionality of the tokenizer with supported models."""

import os
import unittest
from typing import List

import jax
import jax.numpy as jnp
import numpy as np
from sentencepiece import SentencePieceProcessor
from jetstream.engine import tokenizer_pb2, token_utils
from jetstream.engine import engine_api


class SPTokenizer:
  """Tokenier used in original llama2 git"""

  def __init__(self, tokenizer_path: str):
    self.tokenizer = SentencePieceProcessor()
    self.tokenizer.Load(model_file=tokenizer_path)
    assert self.tokenizer.vocab_size() == self.tokenizer.get_piece_size()

  def decode(self, t: List[int]) -> str:
    token = self.tokenizer.decode(t)
    return token


class JetStreamTokenizer:
  """Tokenier used in JetStream before mix_token"""

  def __init__(self, tokenizer_path: str):
    metadata = tokenizer_pb2.TokenizerParameters(path=tokenizer_path)
    self.vocab = token_utils.load_vocab(metadata.path, metadata.extra_ids)

  def decode(self, t: int) -> str:
    token = self.vocab.tokenizer.IdToPiece(t)
    token = token.replace("▁", " ")
    return token


class TokenUtilsTest(unittest.TestCase):

  def setup_sentencepiece(self):
    self.tokenizer_path = "external_tokenizers/llama2/tokenizer.model"
    current_dir = os.path.dirname(__file__)
    self.tokenizer_path = os.path.join(current_dir, self.tokenizer_path)
    print(f"model_path: {self.tokenizer_path}")
    assert os.path.isfile(
        self.tokenizer_path
    ), f"file not found tokenizer_path: {self.tokenizer_path}"
    self.sp_tokenizer = SPTokenizer(self.tokenizer_path)
    self.jt_tokenizer = JetStreamTokenizer(self.tokenizer_path)

  def setup_tiktoken(self):
    self.tokenizer_path = "external_tokenizers/llama3/tokenizer.model"
    current_dir = os.path.dirname(__file__)
    self.tokenizer_path = os.path.join(current_dir, self.tokenizer_path)
    print(f"model_path: {self.tokenizer_path}")
    assert os.path.isfile(
        self.tokenizer_path
    ), f"file not found tokenizer_path: {self.tokenizer_path}"

  def test_decode_vs_piece(self):
    self.setup_sentencepiece()
    tokens = [304, 13, 2266, 526, 777, 9590, 2020, 29901]
    expeted_sp_output = []
    jt_output = []
    for t in tokens:
      expeted_sp_output.append(self.sp_tokenizer.decode([t]))
      jt_output.append(self.jt_tokenizer.decode(t))

    self.assertNotEqual(jt_output, expeted_sp_output)

  def test_sp_vs_seqio(self):
    self.setup_sentencepiece()
    for n in range(0, self.sp_tokenizer.tokenizer.vocab_size()):
      sp_t = self.sp_tokenizer.decode([n])
      seqio_t = self.jt_tokenizer.vocab.tokenizer.decode([n])
      self.assertEqual(sp_t, seqio_t)

  def test_tokenize_and_pad_jax(self):
    jax.config.update("jax_platform_name", "cpu")
    self.setup_sentencepiece()
    s = "I believe the meaning of life is"
    vocab = self.jt_tokenizer.vocab
    max_prefill_length = 1024
    tokens = vocab.encode_tf(s)
    padded_tokens, true_length = token_utils.pad_tokens(
        tokens,
        bos_id=vocab.bos_id,
        pad_id=vocab.pad_id,
        max_prefill_length=max_prefill_length,
    )
    expected_padded_tokens = jnp.array(
        [1, 306, 4658, 278, 6593, 310, 2834, 338, 0, 0, 0, 0, 0, 0, 0, 0]
    )
    expected_true_length = 8
    self.assertTrue(
        jnp.allclose(padded_tokens, expected_padded_tokens, atol=1e-7)
    )
    self.assertEqual(true_length, expected_true_length)

  def test_tokenize_and_pad_np(self):
    self.setup_sentencepiece()
    s = "I believe the meaning of life is"
    vocab = self.jt_tokenizer.vocab
    max_prefill_length = 1024
    tokens = vocab.encode_tf(s)
    padded_tokens, true_length = token_utils.pad_tokens(
        tokens,
        bos_id=vocab.bos_id,
        pad_id=vocab.pad_id,
        max_prefill_length=max_prefill_length,
        jax_padding=False,
    )
    expected_padded_tokens = np.array(
        [1, 306, 4658, 278, 6593, 310, 2834, 338, 0, 0, 0, 0, 0, 0, 0, 0]
    )
    expected_true_length = 8
    self.assertTrue(
        np.allclose(padded_tokens, expected_padded_tokens, atol=1e-7)
    )
    self.assertEqual(true_length, expected_true_length)

  def test_chunk_and_pad_tokens(self):
    jax.config.update("jax_platform_name", "cpu")
    tokens = jnp.arange(0, 65, dtype=jnp.int32)
    _, true_lengths, _ = token_utils.chunk_and_pad_tokens(
        tokens,
        bos_id=1,
        pad_id=0,
        is_bos=True,
        prefill_lengths=[2, 4, 16, 64, 128],
        chunk_size=16,
        max_prefill_length=128,
        jax_padding=True,
    )
    print("true_lengths ", true_lengths)
    assert len(true_lengths) == 5
    assert true_lengths[0] == 17
    assert true_lengths[1] == 16
    assert true_lengths[2] == 16
    assert true_lengths[3] == 16
    assert true_lengths[4] == 1

  def test_tokenize_and_pad(self):
    jax.config.update("jax_platform_name", "cpu")
    self.setup_sentencepiece()
    s = "I believe the meaning of life is"
    vocab = self.jt_tokenizer.vocab
    max_prefill_length = 1024
    padded_tokens, true_length = token_utils.tokenize_and_pad(
        s,
        vocab,
        max_prefill_length=max_prefill_length,
    )
    expected_padded_tokens = jnp.array(
        [1, 306, 4658, 278, 6593, 310, 2834, 338, 0, 0, 0, 0, 0, 0, 0, 0]
    )
    expected_true_length = 8
    self.assertTrue(
        jnp.allclose(padded_tokens, expected_padded_tokens, atol=1e-7)
    )
    self.assertEqual(true_length, expected_true_length)

  def test_pad_token_padding_less_than_zero(self):
    jax.config.update("jax_platform_name", "cpu")
    self.setup_sentencepiece()
    s = "I believe the meaning of life is having different experiences and "
    s += "enjoy everyday of my life."
    vocab = self.jt_tokenizer.vocab
    max_prefill_length = 16
    tokens = vocab.encode_tf(s)
    padded_tokens, true_length = token_utils.pad_tokens(
        tokens,
        bos_id=vocab.bos_id,
        pad_id=vocab.pad_id,
        max_prefill_length=max_prefill_length,
    )
    # Take the last N tokens if we have too many.
    expected_padded_tokens = jnp.array(
        [
            278,
            6593,
            310,
            2834,
            338,
            2534,
            1422,
            27482,
            322,
            13389,
            1432,
            3250,
            310,
            590,
            2834,
            29889,
        ]
    )
    expected_true_length = 19
    self.assertTrue(
        jnp.allclose(padded_tokens, expected_padded_tokens, atol=1e-7)
    )
    self.assertEqual(true_length, expected_true_length)

  def test_sentencepiece_tokenizer_encode(self):
    self.setup_sentencepiece()
    s = "I believe the meaning of life is"
    metadata = tokenizer_pb2.TokenizerParameters(path=self.tokenizer_path)
    tokenizer = token_utils.SentencePieceTokenizer(metadata)
    tokens, true_length = tokenizer.encode(s)
    expected_padded_tokens = np.array(
        [1, 306, 4658, 278, 6593, 310, 2834, 338, 0, 0, 0, 0, 0, 0, 0, 0]
    )
    expected_true_length = 8
    self.assertTrue(np.allclose(tokens, expected_padded_tokens, atol=1e-7))
    self.assertEqual(true_length, expected_true_length)

  def test_sentencepiece_tokenizer_encode_no_bos(self):
    self.setup_sentencepiece()
    s = "I believe the meaning of life is"
    metadata = tokenizer_pb2.TokenizerParameters(path=self.tokenizer_path)
    tokenizer = token_utils.SentencePieceTokenizer(metadata)
    tokens, true_length = tokenizer.encode(s, is_bos=False)
    expected_padded_tokens = np.array(
        [306, 4658, 278, 6593, 310, 2834, 338, 0, 0, 0, 0, 0, 0, 0, 0, 0]
    )
    expected_true_length = 7
    self.assertTrue(np.allclose(tokens, expected_padded_tokens, atol=1e-7))
    self.assertEqual(true_length, expected_true_length)

  def test_sentencepiece_tokenizer_encode_prefill_lengths(self):
    self.setup_sentencepiece()
    s = "I believe the meaning of life is"
    metadata = tokenizer_pb2.TokenizerParameters(path=self.tokenizer_path)
    tokenizer = token_utils.SentencePieceTokenizer(metadata)
    tokens, true_length = tokenizer.encode(s, prefill_lengths=[12, 24, 36])
    expected_padded_tokens = np.array(
        [1, 306, 4658, 278, 6593, 310, 2834, 338, 0, 0, 0, 0]
    )
    expected_true_length = 8
    self.assertTrue(np.allclose(tokens, expected_padded_tokens, atol=1e-7))
    self.assertEqual(true_length, expected_true_length)

  def test_sentencepiece_tokenizer_encode_jax(self):
    jax.config.update("jax_platform_name", "cpu")
    self.setup_sentencepiece()
    s = "I believe the meaning of life is"
    metadata = tokenizer_pb2.TokenizerParameters(path=self.tokenizer_path)
    tokenizer = token_utils.SentencePieceTokenizer(metadata)
    padded_tokens, true_length = tokenizer.encode(s, jax_padding=True)
    expected_padded_tokens = jnp.array(
        [1, 306, 4658, 278, 6593, 310, 2834, 338, 0, 0, 0, 0, 0, 0, 0, 0]
    )
    expected_true_length = 8
    self.assertTrue(
        jnp.allclose(padded_tokens, expected_padded_tokens, atol=1e-7)
    )
    self.assertEqual(true_length, expected_true_length)

  def test_process_result_with_sentencepiece_tokenizer_decode(self):
    self.setup_sentencepiece()
    metadata = tokenizer_pb2.TokenizerParameters(path=self.tokenizer_path)
    tokenizer = token_utils.SentencePieceTokenizer(metadata)
    complete = np.zeros((1,), dtype=np.bool_)

    length = 7
    result_tokens = engine_api.ResultTokens(
        data=np.array(
            [
                [
                    306,
                    4658,
                    278,
                    6593,
                    310,
                    2834,
                    338,
                    1,
                    1,
                    1,
                    1,
                    1,
                    1,
                    1,
                    length,
                ]
            ]
        ),
        tokens_idx=(0, length),
        valid_idx=(length, 2 * length),
        length_idx=(2 * length, 2 * length + 1),
        samples_per_slot=1,
    )
    samples, complete = token_utils.process_result_tokens(
        tokenizer, 0, 16, result_tokens, complete, False
    )
    # Note: the expected_tokens list is for the output token(s) for 1 decode
    # step. Currently, JetStream only output 1 token (1 text piece) for 1
    # decode step.
    expected_tokens = np.array([[306, 4658, 278, 6593, 310, 2834, 338]])
    self.assertTrue(
        np.allclose(
            [sample.token_ids for sample in samples], expected_tokens, atol=1e-7
        )
    )
    self.assertTrue(
        samples[0].text
        == [" I", " believe", " the", " meaning", " of", " life", " is"]
    )
    self.assertTrue(np.allclose(complete, np.zeros((1,), dtype=np.bool_)))

  def test_process_result_with_sentencepiece_tokenizer_client_decode(self):
    self.setup_sentencepiece()
    metadata = tokenizer_pb2.TokenizerParameters(path=self.tokenizer_path)
    tokenizer = token_utils.SentencePieceTokenizer(metadata)
    complete = np.zeros((1,), dtype=np.bool_)

    length = 7
    result_tokens = engine_api.ResultTokens(
        data=np.array(
            [
                [
                    306,
                    4658,
                    278,
                    6593,
                    310,
                    2834,
                    338,
                    1,
                    1,
                    1,
                    1,
                    1,
                    1,
                    1,
                    length,
                ]
            ]
        ),
        tokens_idx=(0, length),
        valid_idx=(length, 2 * length),
        length_idx=(2 * length, 2 * length + 1),
        samples_per_slot=1,
    )
    samples, complete = token_utils.process_result_tokens(
        tokenizer, 0, 16, result_tokens, complete, True
    )
    # Note: the expected_tokens list is for the output token(s) for 1 decode
    # step. Currently, JetStream only output 1 token (1 text piece) for 1
    # decode step.
    expected_tokens = np.array([[306, 4658, 278, 6593, 310, 2834, 338]])
    self.assertTrue(
        np.allclose(
            [sample.token_ids for sample in samples], expected_tokens, atol=1e-7
        )
    )
    # Return token ids only when in client side tokenization mode.
    self.assertTrue(samples[0].text == [])
    self.assertTrue(np.allclose(complete, np.zeros((1,), dtype=np.bool_)))

  def test_sentencepiece_tokenizer_decode(self):
    self.setup_sentencepiece()
    metadata = tokenizer_pb2.TokenizerParameters(path=self.tokenizer_path)
    tokenizer = token_utils.SentencePieceTokenizer(metadata)
    result = tokenizer.decode([306, 4658, 278, 6593, 310, 2834, 338])
    self.assertTrue(result == "I believe the meaning of life is")

  def test_tiktoken_tokenizer_encode(self):
    self.setup_tiktoken()
    s = "I believe the meaning of life is"
    metadata = tokenizer_pb2.TokenizerParameters(path=self.tokenizer_path)
    tokenizer = token_utils.TikToken(metadata)
    tokens, true_length = tokenizer.encode(s)
    expected_padded_tokens = np.array(
        [
            128000,
            40,
            4510,
            279,
            7438,
            315,
            2324,
            374,
            -1,
            -1,
            -1,
            -1,
            -1,
            -1,
            -1,
            -1,
        ]
    )
    expected_true_length = 8
    self.assertTrue(np.allclose(tokens, expected_padded_tokens, atol=1e-7))
    self.assertEqual(true_length, expected_true_length)

  def test_tiktoken_encode_no_bos(self):
    self.setup_tiktoken()
    s = "I believe the meaning of life is"
    metadata = tokenizer_pb2.TokenizerParameters(path=self.tokenizer_path)
    tokenizer = token_utils.TikToken(metadata)
    tokens, true_length = tokenizer.encode(s, is_bos=False)
    expected_padded_tokens = np.array(
        [
            40,
            4510,
            279,
            7438,
            315,
            2324,
            374,
            -1,
            -1,
            -1,
            -1,
            -1,
            -1,
            -1,
            -1,
            -1,
        ]
    )
    expected_true_length = 7
    self.assertTrue(np.allclose(tokens, expected_padded_tokens, atol=1e-7))
    self.assertEqual(true_length, expected_true_length)

  def test_tiktoken_encode_prefill_lengths(self):
    self.setup_tiktoken()
    s = "I believe the meaning of life is"
    metadata = tokenizer_pb2.TokenizerParameters(path=self.tokenizer_path)
    tokenizer = token_utils.TikToken(metadata)
    tokens, true_length = tokenizer.encode(s, prefill_lengths=[12, 24, 36])
    expected_padded_tokens = np.array(
        [128000, 40, 4510, 279, 7438, 315, 2324, 374, -1, -1, -1, -1]
    )
    expected_true_length = 8
    self.assertTrue(np.allclose(tokens, expected_padded_tokens, atol=1e-7))
    self.assertEqual(true_length, expected_true_length)

  def test_tiktoken_encode_jax(self):
    jax.config.update("jax_platform_name", "cpu")
    self.setup_tiktoken()
    s = "I believe the meaning of life is"
    metadata = tokenizer_pb2.TokenizerParameters(path=self.tokenizer_path)
    tokenizer = token_utils.TikToken(metadata)
    padded_tokens, true_length = tokenizer.encode(s, jax_padding=True)
    expected_padded_tokens = jnp.array(
        [
            128000,
            40,
            4510,
            279,
            7438,
            315,
            2324,
            374,
            -1,
            -1,
            -1,
            -1,
            -1,
            -1,
            -1,
            -1,
        ]
    )
    expected_true_length = 8
    self.assertTrue(
        jnp.allclose(padded_tokens, expected_padded_tokens, atol=1e-7)
    )
    self.assertEqual(true_length, expected_true_length)

  def test_process_result_with_tiktoken_decode(self):
    self.setup_tiktoken()
    metadata = tokenizer_pb2.TokenizerParameters(path=self.tokenizer_path)
    tokenizer = token_utils.TikToken(metadata)
    complete = np.zeros((1,), dtype=np.bool_)

    length = 7
    result_tokens = engine_api.ResultTokens(
        data=np.array(
            [[40, 4510, 279, 7438, 315, 2324, 374, 1, 1, 1, 1, 1, 1, 1, length]]
        ),
        tokens_idx=(0, length),
        valid_idx=(length, 2 * length),
        length_idx=(2 * length, 2 * length + 1),
        samples_per_slot=1,
    )
    samples, complete = token_utils.process_result_tokens(
        tokenizer, 0, 16, result_tokens, complete, False
    )
    # Note: the expected_tokens list is for the output token(s) for 1 decode
    # step. Currently, JetStream only output 1 token (1 text piece) for 1
    # decode step.
    expected_tokens = np.array([[40, 4510, 279, 7438, 315, 2324, 374]])
    self.assertTrue(
        np.allclose(
            [sample.token_ids for sample in samples], expected_tokens, atol=1e-7
        )
    )
    self.assertTrue(
        samples[0].text
        == ["I", " believe", " the", " meaning", " of", " life", " is"]
    )
    self.assertTrue(np.allclose(complete, np.zeros((1,), dtype=np.bool_)))

  def test_process_result_with_tiktoken_client_decode(self):
    self.setup_tiktoken()
    metadata = tokenizer_pb2.TokenizerParameters(path=self.tokenizer_path)
    tokenizer = token_utils.TikToken(metadata)
    complete = np.zeros((1,), dtype=np.bool_)

    length = 7
    result_tokens = engine_api.ResultTokens(
        data=np.array(
            [[40, 4510, 279, 7438, 315, 2324, 374, 1, 1, 1, 1, 1, 1, 1, length]]
        ),
        tokens_idx=(0, length),
        valid_idx=(length, 2 * length),
        length_idx=(2 * length, 2 * length + 1),
        samples_per_slot=1,
    )
    samples, complete = token_utils.process_result_tokens(
        tokenizer, 0, 16, result_tokens, complete, True
    )
    # Note: the expected_tokens list is for the output token(s) for 1 decode
    # step. Currently, JetStream only output 1 token (1 text piece) for 1
    # decode step.
    expected_tokens = np.array([[40, 4510, 279, 7438, 315, 2324, 374]])
    self.assertTrue(
        np.allclose(
            [sample.token_ids for sample in samples], expected_tokens, atol=1e-7
        )
    )
    # Return token ids only when in client side tokenization mode.
    self.assertTrue(samples[0].text == [])
    self.assertTrue(np.allclose(complete, np.zeros((1,), dtype=np.bool_)))

  def test_tiktoken_decode(self):
    self.setup_tiktoken()
    metadata = tokenizer_pb2.TokenizerParameters(path=self.tokenizer_path)
    tokenizer = token_utils.TikToken(metadata)
    result = tokenizer.decode([40, 4510, 279, 7438, 315, 2324, 374])
    self.assertTrue(result == "I believe the meaning of life is")

  def test_text_tokens_to_str(self):
    # Start with text token
    self.assertTrue(
        token_utils.text_tokens_to_str(
            ["你", "好", "<0xE5>", "<0x90>", "<0x97>", "hello"]
        )
        == "你好吗hello"
    )
    self.assertTrue(
        token_utils.text_tokens_to_str(
            ["你", "好", "<0xE5>", "<0x90>", "<0x97>", "<0x0A>", "hello"]
        )
        == "你好吗\nhello"
    )
    self.assertTrue(
        token_utils.text_tokens_to_str(
            [
                "你",
                "好",
                "<0xE5>",
                "<0x90>",
                "<0x97>",
                "<0x0A>",
                "<0x0A>",
                "hello",
            ]
        )
        == "你好吗\n\nhello"
    )
    self.assertTrue(
        token_utils.text_tokens_to_str(
            ["你", "好", "<0xE5>", "<0x90>", "<0x97>", "hello", "<0x0A>"]
        )
        == "你好吗hello\n"
    )
    # Start with byte token
    self.assertTrue(
        token_utils.text_tokens_to_str(
            ["<0x0A>", "你", "好", "<0xE5>", "<0x90>", "<0x97>", "hello"]
        )
        == "\n你好吗hello"
    )
    self.assertTrue(
        token_utils.text_tokens_to_str(
            [
                "<0x0A>",
                "<0x0A>",
                "你",
                "好",
                "<0xE5>",
                "<0x90>",
                "<0x97>",
                "hello",
            ]
        )
        == "\n\n你好吗hello"
    )
    self.assertTrue(
        token_utils.text_tokens_to_str(["<0xE5>", "<0x90>", "<0x97>", "hello"])
        == "吗hello"
    )
    self.assertTrue(
        token_utils.text_tokens_to_str(
            ["<0x0A>", "<0x0A>", "<0xE5>", "<0x90>", "<0x97>", "hello"]
        )
        == "\n\n吗hello"
    )
    self.assertTrue(
        token_utils.text_tokens_to_str(
            ["<0x0A>", "<0x0A>", "<0xE5>", "<0x90>", "<0x97>"]
        )
        == "\n\n吗"
    )
    # Invalid byte token sequence
    self.assertTrue(
        token_utils.text_tokens_to_str(
            ["你", "好", "<0xE5>", "<0x90>", "<0x0A>", "<0x97>", "hello"]
        )
        == "你好�\n�hello"
    )



#  Copyright 2025 Google LLC
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#       https://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

"""
Runs SFT trainer correctness with TRL implementation.

Usage:
python3 MaxText/tests/sft_trainer_correctness.py --model-name=llama3.1-8b --tokenizer-path=meta-llama/Llama-3.1-8B --model-ckpt-path=gs://maxtext-model-checkpoints/llama3.1-8b/2025-01-23-19-04/scanned/0/items
"""

import argparse
import jax
import jax.numpy as jnp
import jsonlines
import numpy as np
import os
import sys

from jax.sharding import Mesh
from transformers import AutoTokenizer

current_dir = os.path.dirname(os.path.abspath(__file__))
maxtext_parent_dir = os.path.dirname(current_dir)
sys.path.append(maxtext_parent_dir)

import max_logging
import max_utils
import pyconfig
from input_pipeline import _input_pipeline_utils
from layers import models
from layers import quantizations


def initialize_config(config):
  return pyconfig.initialize(
      [sys.argv[0], "MaxText/configs/sft.yml"],
      run_name="test-sft-trainer-correctness",
      model_name=config.model_name,
      tokenizer_path=config.tokenizer_path,
      enable_checkpointing=True,
      load_parameters_path=config.model_ckpt_path,
      max_target_length=config.max_target_length,
      per_device_batch_size=1,
      max_prefill_predict_length=32,
      dataset_type="synthetic",
      dtype="float32",
      matmul_precision="high",
      logits_dot_in_fp32=True,
      skip_jax_distributed_system=True,
  )


def get_golden_data(config):
  """Get the golden data for SFTTrainer in TRL."""
  golden_data_path = "MaxText/test_assets/golden_data_sft_" + config.model_name + ".jsonl"
  with jsonlines.open(golden_data_path, "r") as f:
    golden_data = list(f)
  return golden_data[0]


def setup_maxtext_model(config):
  init_rng = jax.random.PRNGKey(config.init_weights_seed)
  init_rng, rng1 = jax.random.split(init_rng)
  quant = quantizations.configure_quantization(config)
  devices_array = max_utils.create_device_mesh(config)
  mesh = Mesh(devices_array, config.mesh_axes)
  maxtext_model = models.Transformer(config=config, mesh=mesh, quant=quant)
  state, _ = max_utils.setup_decode_state(maxtext_model, config, rng1, mesh, None)
  return maxtext_model, state, init_rng


def prepare_maxtext_inputs(maxtext_data, config):
  data = _input_pipeline_utils.extract_messages_and_mask(maxtext_data, "messages")
  tokenizer = AutoTokenizer.from_pretrained(
      config.tokenizer_path,
      add_bos_token=False,
      add_eos_token=False,
      model_max_length=config.max_target_length,
  )
  tokenized_data = _input_pipeline_utils.tokenization(
      data,
      hf_tokenizer=tokenizer,
      truncation=False,
      max_length=config.max_target_length,
      column_names=["messages"],
  )
  masked_inputs = _input_pipeline_utils.SFTPromptMasking(
      text_column_name="messages",
      completion_only=False,
      max_target_length=config.max_target_length,
      add_bos=True,
      add_eos=True,
      bos_id=tokenizer.bos_token_id,
      eos_id=tokenizer.eos_token_id,
      unk_id=tokenizer.unk_token_id,
  ).map(tokenized_data)

  global_batch_size = int(jax.device_count() * config.per_device_batch_size * config.gradient_accumulation_steps)
  inputs = jnp.stack([np.asarray(masked_inputs["inputs"], dtype=np.int32) for _ in range(global_batch_size)])
  inputs_segmentation = jnp.stack([(masked_inputs["inputs"] != 0).astype(np.int32) for _ in range(global_batch_size)])
  inputs_position = jnp.stack(
      [np.arange(masked_inputs["inputs"].shape[0], dtype=np.int32) for _ in range(global_batch_size)]
  )
  return inputs, inputs_segmentation, inputs_position


def get_maxtext_logits(inputs, inputs_position, inputs_segmentation, config):
  maxtext_model, state, rng = setup_maxtext_model(config)
  maxtext_logits, _ = maxtext_model.apply(
      state.params,
      inputs,
      inputs_position,
      decoder_segment_ids=inputs_segmentation,
      enable_dropout=False,
      rngs={"aqt": rng},
      mutable="intermediates",
  )
  return maxtext_logits


def get_kl_div(maxtext_logits, hf_logits):
  maxtext_probabilities = jax.nn.softmax(maxtext_logits, axis=-1)
  hf_probabilities = jax.nn.softmax(hf_logits, axis=-1)
  kl_div = jax.numpy.sum(jax.scipy.special.kl_div(hf_probabilities, maxtext_probabilities), axis=-1)
  return kl_div


def main(config):
  golden_data = get_golden_data(config)
  inputs, inputs_segmentation, inputs_position = prepare_maxtext_inputs(golden_data["data"], config)
  maxtext_data = {
      "inputs": inputs,
      "inputs_segmentation": inputs_segmentation,
      "inputs_position": inputs_position,
  }
  maxtext_logits = get_maxtext_logits(inputs, inputs_position, inputs_segmentation, config)

  assert golden_data["tokens"] == maxtext_data["inputs"][0].tolist()
  assert golden_data["attention_mask"] == maxtext_data["inputs_segmentation"][0].tolist()

  trl_logits = np.array(golden_data["trl_logits"])

  max_logging.log(f"Max numerical difference: {np.max(np.abs(np.subtract(maxtext_logits[0], trl_logits)))}")
  assert jax.numpy.allclose(
      maxtext_logits[0],
      trl_logits,
      rtol=1e-05,
      atol=0.06,
      equal_nan=False,
  )

  kl_div = get_kl_div(maxtext_logits, trl_logits)
  max_logging.log(f"KL divergence: {kl_div}, max KL divergence: {jnp.max(kl_div)}")
  assert jax.numpy.all(kl_div < 7e-5)


if __name__ == "__main__":
  parser = argparse.ArgumentParser()
  parser.add_argument("--model-name", type=str, required=True)
  parser.add_argument("--tokenizer-path", type=str, required=True)
  parser.add_argument("--model-ckpt-path", type=str, required=True)
  parser.add_argument("--max-target-length", type=int, required=False, default=64)
  parser_config = parser.parse_args(sys.argv[1:])
  config = initialize_config(parser_config)
  main(config)



"""
Copyright 2024 Google LLC

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

     https://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""

""" collective matmul for linear layer (1d ring topology).

Reference:
https://dl.acm.org/doi/pdf/10.1145/3567955.3567959

The implementation is throughput-bound. Latency-bound implementation
should also be added.

The implementation is via the JAX/XLA collective-permute API, and it doesn't
work for GPU well: https://github.com/openxla/xla/issues/10640#issuecomment-2246448416
"""

import jax
from jax import numpy as jnp
from jax.sharding import Mesh
from jax.experimental.shard_map import shard_map
from inference.parallel import get_num_partitions, get_partition_index, tp_axis_names


def prepare_rhs_for_all_gather_collective_matmul(rhs: jax.Array, mesh: Mesh):
  """Prepare rhs for all gather collective matmul.

  For bidirectional collective matmul, the two lhs chunks
  received from the neighbor are not contiguous in the
  all-gathered lhs. Reshuffle the rhs ahead to process
  the non-contiguous lhs chunks together in the collective
  matmul.
  """
  axis_names = tp_axis_names()

  def reshuffle(rhs: jax.Array):
    idx = get_partition_index(axis_names=axis_names)
    num_partitions = get_num_partitions(axis_names=axis_names)
    rhs_chunk_row_size = rhs.shape[0] // num_partitions
    half_rhs_chunk_row_size = rhs_chunk_row_size // 2

    def swap(i, carry):
      rhs = carry
      idx_1 = ((idx + i) % num_partitions) * rhs_chunk_row_size
      idx_2 = ((idx - i) % num_partitions) * rhs_chunk_row_size
      operand_1 = jax.lax.dynamic_slice_in_dim(
          rhs, idx_1, half_rhs_chunk_row_size, axis=0
      )
      operand_2 = jax.lax.dynamic_slice_in_dim(
          rhs, idx_2, half_rhs_chunk_row_size, axis=0
      )
      rhs = jax.lax.dynamic_update_slice_in_dim(rhs, operand_1, idx_2, axis=0)
      rhs = jax.lax.dynamic_update_slice_in_dim(rhs, operand_2, idx_1, axis=0)
      return rhs

    rhs = jax.lax.fori_loop(1, num_partitions // 2, swap, rhs)
    return rhs

  return shard_map(
      f=reshuffle,
      mesh=mesh,
      in_specs=rhs.sharding.spec,
      out_specs=rhs.sharding.spec,
  )(rhs)


def all_gather_collective_matmul(lhs, rhs, axis_names):
  """All gather collective matmul.

  The function works for matmul where the lhs is partitioned at
  the contracting dimension.
  """
  idx = get_partition_index(axis_names=axis_names)
  num_partitions = get_num_partitions(axis_names=axis_names)
  rhs_chunk_row_size = rhs.shape[0] // num_partitions

  def step(i, carry):
    accum, fwd_lhs, bwd_lhs = carry
    rhs_row_idx = ((idx + i) % num_partitions) * rhs_chunk_row_size
    cur_lhs = jnp.concatenate((fwd_lhs, bwd_lhs), axis=1)
    rhs_chunk = jax.lax.dynamic_slice_in_dim(
        rhs, rhs_row_idx, rhs_chunk_row_size
    )
    output = cur_lhs @ rhs_chunk
    accum += output
    fwd_lhs = jax.lax.ppermute(
        fwd_lhs,
        axis_names,
        [(j, (j + 1) % num_partitions) for j in range(num_partitions)],
    )
    bwd_lhs = jax.lax.ppermute(
        bwd_lhs,
        axis_names,
        [(j, (j - 1) % num_partitions) for j in range(num_partitions)],
    )
    return accum, fwd_lhs, bwd_lhs

  res_shape = (lhs.shape[0], rhs.shape[1])
  accum = jnp.zeros(shape=res_shape, dtype=lhs.dtype)
  fwd_lhs, bwd_lhs = jnp.split(lhs, 2, 1)
  accum, fwd_lhs, bwd_lhs = jax.lax.fori_loop(
      0, num_partitions - 1, step, (accum, fwd_lhs, bwd_lhs)
  )

  # Last round which doesn't need collective permute.
  rhs_row_idx = (
      (idx + num_partitions - 1) % num_partitions
  ) * rhs_chunk_row_size
  cur_lhs = jnp.concatenate((fwd_lhs, bwd_lhs), axis=1)
  rhs_chunk = jax.lax.dynamic_slice_in_dim(rhs, rhs_row_idx, rhs_chunk_row_size)
  output = cur_lhs @ rhs_chunk
  accum += output

  return accum


def prepare_rhs_for_collective_matmul_reduce_scatter(
    rhs: jax.Array, mesh: Mesh
):
  """Prepare rhs for collective matmul with reduce scatter.

  For bidirectional collective matmul, the two accum chunks
  received from the neighbor are not contiguous in the
  final accum. Reshuffle the rhs ahead to process
  the non-contiguous accum chunks together in the collective
  matmul.
  """
  axis_names = tp_axis_names()

  def reshuffle(rhs):
    idx = get_partition_index(axis_names=axis_names)
    num_partitions = get_num_partitions(axis_names=axis_names)
    rhs_chunk_col_size = rhs.shape[1] // num_partitions
    half_rhs_chunk_col_size = rhs_chunk_col_size // 2

    def swap(i, carry):
      rhs = carry
      idx_1 = ((idx + i) % num_partitions) * rhs_chunk_col_size
      idx_2 = ((idx - i) % num_partitions) * rhs_chunk_col_size
      operand_1 = jax.lax.dynamic_slice_in_dim(
          rhs, idx_1, half_rhs_chunk_col_size, axis=1
      )
      operand_2 = jax.lax.dynamic_slice_in_dim(
          rhs, idx_2, half_rhs_chunk_col_size, axis=1
      )
      rhs = jax.lax.dynamic_update_slice_in_dim(rhs, operand_1, idx_2, axis=1)
      rhs = jax.lax.dynamic_update_slice_in_dim(rhs, operand_2, idx_1, axis=1)
      return rhs

    rhs = jax.lax.fori_loop(1, num_partitions // 2, swap, rhs)
    return rhs

  pspec = rhs.sharding.spec

  return shard_map(
      f=reshuffle,
      mesh=mesh,
      in_specs=pspec,
      out_specs=pspec,
  )(rhs)


def collective_matmul_reduce_scatter(lhs, rhs, axis_names):
  """Collective matmul with reduce scatter at the output column axis."""
  idx = get_partition_index(axis_names=axis_names)
  num_partitions = get_num_partitions(axis_names=axis_names)
  rhs_chunk_col_size = rhs.shape[1] // num_partitions
  # Compute the partial result for the chip at the last step.
  rhs_col_idx = ((idx + 1) % num_partitions) * rhs_chunk_col_size
  rhs_chunk = jax.lax.dynamic_slice_in_dim(
      rhs,
      start_index=rhs_col_idx,
      slice_size=rhs_chunk_col_size,
      axis=1,
  )
  partial_res = lhs @ rhs_chunk
  accum_to_send_shape = (lhs.shape[0], rhs_chunk_col_size // 2)
  fwd_accum = jnp.zeros(shape=accum_to_send_shape)
  bwd_accum = jnp.zeros(shape=accum_to_send_shape)

  def step(i, carry):
    fwd_accum, bwd_accum, partial_res = carry
    accum = jnp.concatenate((fwd_accum, bwd_accum), axis=1)
    accum += partial_res
    fwd_accum, bwd_accum = jnp.split(accum, 2, axis=1)

    rhs_col_idx = ((idx + 1 + i) % num_partitions) * rhs_chunk_col_size
    rhs_chunk = jax.lax.dynamic_slice_in_dim(
        rhs,
        start_index=rhs_col_idx,
        slice_size=rhs_chunk_col_size,
        axis=1,
    )
    partial_res = lhs @ rhs_chunk
    fwd_accum = jax.lax.ppermute(
        fwd_accum,
        axis_names,
        [(j, (j + 1) % num_partitions) for j in range(num_partitions)],
    )
    bwd_accum = jax.lax.ppermute(
        bwd_accum,
        axis_names,
        [(j, (j - 1) % num_partitions) for j in range(num_partitions)],
    )
    return fwd_accum, bwd_accum, partial_res

  fwd_accum, bwd_accum, partial_res = jax.lax.fori_loop(
      1, num_partitions, step, (fwd_accum, bwd_accum, partial_res)
  )
  accum = jnp.concatenate((fwd_accum, bwd_accum), axis=1)
  accum += partial_res
  return accum



# Copyright 2024 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Http API server protocol."""

from pydantic import BaseModel  # type: ignore


class TextContent(BaseModel):
  text: str


class TokenContent(BaseModel):
  token_ids: list[int]


class Metadata(BaseModel):
  start_time: float


class DecodeRequest(BaseModel):
  max_tokens: int
  text_content: TextContent | None = None
  token_content: TokenContent | None = None
  metadata: Metadata | None = None

  # Config to enforce the oneof behavior at runtime.
  class Config:
    extra = "forbid"  # Prevent extra fields.
    anystr_strip_whitespace = True
